# pylint: skip-file
import inspect
import math
from dataclasses import dataclass

import torch
from torch import nn
from torch.nn import functional as F


class LayerNorm(nn.Module):
    """LayerNorm but with an optional bias. PyTorch doesn't support simply bias=False"""

    def __init__(self, ndim, bias):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(ndim))
        self.bias = nn.Parameter(torch.zeros(ndim)) if bias else None

    def forward(self, input):
        return F.layer_norm(input, self.weight.shape, self.weight, self.bias, 1e-5)


class CausalSelfAttention(nn.Module):
    def __init__(self, config):
        super().__init__()
        assert config.n_embd % config.n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(config.n_embd, 3 * config.n_embd, bias=config.bias)
        # output projection
        self.c_proj = nn.Linear(config.n_embd, config.n_embd, bias=config.bias)
        # regularization
        self.attn_dropout = nn.Dropout(config.dropout)
        self.resid_dropout = nn.Dropout(config.dropout)
        self.n_head = config.n_head
        self.n_embd = config.n_embd
        self.dropout = config.dropout
        # flash attention make GPU go brrrrr but support is only in PyTorch >= 2.0
        self.flash = hasattr(torch.nn.functional, "scaled_dot_product_attention")
        if not self.flash:
            print(
                "WARNING: using slow attention. Flash Attention requires PyTorch >= 2.0"
            )
            # causal mask to ensure that attention is only applied to the left in the input sequence
            self.register_buffer(
                "bias",
                torch.tril(torch.ones(config.block_size, config.block_size)).view(
                    1, 1, config.block_size, config.block_size
                ),
            )

    def forward(self, x):
        (
            B,
            T,
            C,
        ) = x.size()  # batch size, sequence length, embedding dimensionality (n_embd)

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        if self.flash:
            # efficient attention using Flash Attention CUDA kernels
            y = torch.nn.functional.scaled_dot_product_attention(
                q,
                k,
                v,
                attn_mask=None,
                dropout_p=self.dropout if self.training else 0,
                is_causal=True,
            )
        else:
            # manual implementation of attention
            att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
            att = att.masked_fill(self.bias[:, :, :T, :T] == 0, float("-inf"))
            att = F.softmax(att, dim=-1)
            att = self.attn_dropout(att)
            y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = (
            y.transpose(1, 2).contiguous().view(B, T, C)
        )  # re-assemble all head outputs side by side

        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y


class MLP(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.c_fc = nn.Linear(config.n_embd, 4 * config.n_embd, bias=config.bias)
        self.gelu = nn.GELU()
        self.c_proj = nn.Linear(4 * config.n_embd, config.n_embd, bias=config.bias)
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, x):
        x = self.c_fc(x)
        x = self.gelu(x)
        x = self.c_proj(x)
        x = self.dropout(x)
        return x


class Block(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.ln_1 = LayerNorm(config.n_embd, bias=config.bias)
        self.attn = CausalSelfAttention(config)
        self.ln_2 = LayerNorm(config.n_embd, bias=config.bias)
        self.mlp = MLP(config)

    def forward(self, x):
        x = x + self.attn(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


@dataclass
class GPTConfig:
    block_size: int = 1024
    vocab_size: int = 50304  # GPT-2 vocab_size of 50257, padded up to nearest multiple of 64 for efficiency
    n_layer: int = 12
    n_head: int = 12
    n_embd: int = 768
    dropout: float = 0.0
    bias: bool = True  # True: bias in Linears and LayerNorms, like GPT-2. False: a bit better and faster


class GPT(nn.Module):
    def __init__(self, config: GPTConfig):
        super().__init__()
        assert config.vocab_size is not None
        assert config.block_size is not None
        self.config = config

        self.transformer = nn.ModuleDict(
            dict(
                wte=nn.Embedding(config.vocab_size, config.n_embd),
                wpe=nn.Embedding(config.block_size, config.n_embd),
                drop=nn.Dropout(config.dropout),
                h=nn.ModuleList([Block(config) for _ in range(config.n_layer)]),
                ln_f=LayerNorm(config.n_embd, bias=config.bias),
            )
        )
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size, bias=False)
        # with weight tying when using torch.compile() some warnings get generated:
        # "UserWarning: functional_call was passed multiple values for tied weights.
        # This behavior is deprecated and will be an error in future versions"
        # not 100% sure what this is, so far seems to be harmless. TODO investigate
        self.transformer.wte.weight = (
            self.lm_head.weight
        )  # https://paperswithcode.com/method/weight-tying

        # init all weights
        self.apply(self._init_weights)
        # apply special scaled init to the residual projections, per GPT-2 paper
        for pn, p in self.named_parameters():
            if pn.endswith("c_proj.weight"):
                torch.nn.init.normal_(
                    p, mean=0.0, std=0.02 / math.sqrt(2 * config.n_layer)
                )

        # report number of parameters
        print("number of parameters: %.2fM" % (self.get_num_params() / 1e6,))

    def get_num_params(self, non_embedding=True):
        """
        Return the number of parameters in the model.
        For non-embedding count (default), the position embeddings get subtracted.
        The token embeddings would too, except due to the parameter sharing these
        params are actually used as weights in the final layer, so we include them.
        """
        n_params = sum(p.numel() for p in self.parameters())
        if non_embedding:
            n_params -= self.transformer.wpe.weight.numel()
        return n_params

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

    def forward(self, idx, targets=None):
        device = idx.device
        b, t = idx.size()
        assert (
            t <= self.config.block_size
        ), f"Cannot forward sequence of length {t}, block size is only {self.config.block_size}"
        pos = torch.arange(0, t, dtype=torch.long, device=device)  # shape (t)

        # forward the GPT model itself
        tok_emb = self.transformer.wte(idx)  # token embeddings of shape (b, t, n_embd)
        pos_emb = self.transformer.wpe(pos)  # position embeddings of shape (t, n_embd)
        x = self.transformer.drop(tok_emb + pos_emb)
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)

        if targets is not None:
            # if we are given some desired targets also calculate the loss
            logits = self.lm_head(x)
            loss = F.cross_entropy(
                logits.view(-1, logits.size(-1)), targets.view(-1), ignore_index=-1
            )
        else:
            # inference-time mini-optimization: only forward the lm_head on the very last position
            logits = self.lm_head(
                x[:, [-1], :]
            )  # note: using list [-1] to preserve the time dim
            loss = None

        return logits, loss

    def crop_block_size(self, block_size):
        # model surgery to decrease the block size if necessary
        # e.g. we may load the GPT2 pretrained model checkpoint (block size 1024)
        # but want to use a smaller block size for some smaller, simpler model
        assert block_size <= self.config.block_size
        self.config.block_size = block_size
        self.transformer.wpe.weight = nn.Parameter(
            self.transformer.wpe.weight[:block_size]
        )
        for block in self.transformer.h:
            if hasattr(block.attn, "bias"):
                block.attn.bias = block.attn.bias[:, :, :block_size, :block_size]

    @classmethod
    def from_pretrained(cls, model_type, override_args=None):
        assert model_type in {"gpt2", "gpt2-medium", "gpt2-large", "gpt2-xl"}
        override_args = override_args or {}  # default to empty dict
        # only dropout can be overridden see more notes below
        assert all(k == "dropout" for k in override_args)
        from transformers import GPT2LMHeadModel

        print("loading weights from pretrained gpt: %s" % model_type)

        # n_layer, n_head and n_embd are determined from model_type
        config_args = {
            "gpt2": dict(n_layer=12, n_head=12, n_embd=768),  # 124M params
            "gpt2-medium": dict(n_layer=24, n_head=16, n_embd=1024),  # 350M params
            "gpt2-large": dict(n_layer=36, n_head=20, n_embd=1280),  # 774M params
            "gpt2-xl": dict(n_layer=48, n_head=25, n_embd=1600),  # 1558M params
        }[model_type]
        print("forcing vocab_size=50257, block_size=1024, bias=True")
        config_args["vocab_size"] = 50257  # always 50257 for GPT model checkpoints
        config_args["block_size"] = 1024  # always 1024 for GPT model checkpoints
        config_args["bias"] = True  # always True for GPT model checkpoints
        # we can override the dropout rate, if desired
        if "dropout" in override_args:
            print(f"overriding dropout rate to {override_args['dropout']}")
            config_args["dropout"] = override_args["dropout"]
        # create a from-scratch initialized minGPT model
        config = GPTConfig(**config_args)
        model = GPT(config)
        sd = model.state_dict()
        sd_keys = sd.keys()
        sd_keys = [
            k for k in sd_keys if not k.endswith(".attn.bias")
        ]  # discard this mask / buffer, not a param

        # init a huggingface/transformers model
        model_hf = GPT2LMHeadModel.from_pretrained(model_type)
        sd_hf = model_hf.state_dict()

        # copy while ensuring all of the parameters are aligned and match in names and shapes
        sd_keys_hf = sd_hf.keys()
        sd_keys_hf = [
            k for k in sd_keys_hf if not k.endswith(".attn.masked_bias")
        ]  # ignore these, just a buffer
        sd_keys_hf = [
            k for k in sd_keys_hf if not k.endswith(".attn.bias")
        ]  # same, just the mask (buffer)
        transposed = [
            "attn.c_attn.weight",
            "attn.c_proj.weight",
            "mlp.c_fc.weight",
            "mlp.c_proj.weight",
        ]
        # basically the openai checkpoints use a "Conv1D" module, but we only want to use a vanilla Linear
        # this means that we have to transpose these weights when we import them
        assert len(sd_keys_hf) == len(
            sd_keys
        ), f"mismatched keys: {len(sd_keys_hf)} != {len(sd_keys)}"
        for k in sd_keys_hf:
            if any(k.endswith(w) for w in transposed):
                # special treatment for the Conv1D weights we need to transpose
                assert sd_hf[k].shape[::-1] == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[k].t())
            else:
                # vanilla copy over the other parameters
                assert sd_hf[k].shape == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[k])

        return model

    def configure_optimizers(self, weight_decay, learning_rate, betas, device_type):
        # start with all of the candidate parameters
        param_dict = {pn: p for pn, p in self.named_parameters()}
        # filter out those that do not require grad
        param_dict = {pn: p for pn, p in param_dict.items() if p.requires_grad}
        # create optim groups. Any parameters that is 2D will be weight decayed, otherwise no.
        # i.e. all weight tensors in matmuls + embeddings decay, all biases and layernorms don't.
        decay_params = [p for n, p in param_dict.items() if p.dim() >= 2]
        nodecay_params = [p for n, p in param_dict.items() if p.dim() < 2]
        optim_groups = [
            {"params": decay_params, "weight_decay": weight_decay},
            {"params": nodecay_params, "weight_decay": 0.0},
        ]
        num_decay_params = sum(p.numel() for p in decay_params)
        num_nodecay_params = sum(p.numel() for p in nodecay_params)
        print(
            f"num decayed parameter tensors: {len(decay_params)}, with {num_decay_params:,} parameters"
        )
        print(
            f"num non-decayed parameter tensors: {len(nodecay_params)}, with {num_nodecay_params:,} parameters"
        )
        # Create AdamW optimizer and use the fused version if it is available
        fused_available = "fused" in inspect.signature(torch.optim.AdamW).parameters
        use_fused = fused_available and device_type == "cuda"
        extra_args = dict(fused=True) if use_fused else dict()
        optimizer = torch.optim.AdamW(
            optim_groups, lr=learning_rate, betas=betas, **extra_args
        )
        print(f"using fused AdamW: {use_fused}")

        return optimizer

    def estimate_mfu(self, fwdbwd_per_iter, dt):
        """estimate model flops utilization (MFU) in units of A100 bfloat16 peak FLOPS"""
        # first estimate the number of flops we do per iteration.
        # see PaLM paper Appendix B as ref: https://arxiv.org/abs/2204.02311
        N = self.get_num_params()
        cfg = self.config
        L, H, Q, T = cfg.n_layer, cfg.n_head, cfg.n_embd // cfg.n_head, cfg.block_size
        flops_per_token = 6 * N + 12 * L * H * Q * T
        flops_per_fwdbwd = flops_per_token * T
        flops_per_iter = flops_per_fwdbwd * fwdbwd_per_iter
        # express our flops throughput as ratio of A100 bfloat16 peak flops
        flops_achieved = flops_per_iter * (1.0 / dt)  # per second
        flops_promised = 312e12  # A100 GPU bfloat16 peak flops is 312 TFLOPS
        mfu = flops_achieved / flops_promised
        return mfu

    @torch.no_grad()
    def generate(self, idx, max_new_tokens, temperature=1.0, top_k=None):
        """
        Take a conditioning sequence of indices idx (LongTensor of shape (b,t)) and complete
        the sequence max_new_tokens times, feeding the predictions back into the model each time.
        Most likely you'll want to make sure to be in model.eval() mode of operation for this.
        """
        for _ in range(max_new_tokens):
            # if the sequence context is growing too long we must crop it at block_size
            idx_cond = (
                idx
                if idx.size(1) <= self.config.block_size
                else idx[:, -self.config.block_size :]
            )
            # forward the model to get the logits for the index in the sequence
            logits, _ = self(idx_cond)
            # pluck the logits at the final step and scale by desired temperature
            logits = logits[:, -1, :] / temperature
            # optionally crop the logits to only the top k options
            if top_k is not None:
                v, _ = torch.topk(logits, min(top_k, logits.size(-1)))
                logits[logits < v[:, [-1]]] = -float("Inf")
            # apply softmax to convert logits to (normalized) probabilities
            probs = F.softmax(logits, dim=-1)
            # sample from the distribution
            idx_next = torch.multinomial(probs, num_samples=1)
            # append sampled index to the running sequence and continue
            idx = torch.cat((idx, idx_next), dim=1)

        return idx



from abc import ABCMeta, abstractmethod


class BaseStorageBackend(metaclass=ABCMeta):
    """Abstract class of storage backends.

    All backends need to implement two apis: :meth:`get()` and
    :meth:`get_text()`.

    - :meth:`get()` reads the file as a byte stream.
    - :meth:`get_text()` reads the file as texts.
    """

    # a flag to indicate whether the backend can create a symlink for a file
    # This attribute will be deprecated in future.
    _allow_symlink = False

    @property
    def allow_symlink(self):
        return self._allow_symlink

    @property
    def name(self):
        return self.__class__.__name__

    @abstractmethod
    def get(self, filepath):
        pass

    @abstractmethod
    def get_text(self, filepath):
        pass



from typing import Union

import torch
from torch import distributions


class BoxUniform(distributions.Independent):
    def __init__(
        self,
        low: Union[torch.Tensor, float],
        high: Union[torch.Tensor, float],
        reinterpreted_batch_ndims: int = 1,
    ):
        """Multidimensionqal uniform distribution defined on a box.
        
        A `Uniform` distribution initialized with e.g. a parameter vector low or high of length 3 will result in a /batch/ dimension of length 3. A log_prob evaluation will then output three numbers, one for each of the independent Uniforms in the batch. Instead, a `BoxUniform` initialized in the same way has three /event/ dimensions, and returns a scalar log_prob corresponding to whether the evaluated point is in the box defined by low and high or outside. 
    
        Refer to torch.distributions.Uniform and torch.distributions.Independent for further documentation.
    
        Args:
            low (Tensor or float): lower range (inclusive).
            high (Tensor or float): upper range (exclusive).
            reinterpreted_batch_ndims (int): the number of batch dims to
                                             reinterpret as event dims.
        """

        super().__init__(
            distributions.Uniform(low=low, high=high), reinterpreted_batch_ndims
        )


class MG1Uniform(distributions.Uniform):
    def log_prob(self, value):
        return super().log_prob(self._to_noise(value))

    def sample(self, sample_shape=torch.Size()):
        return self._to_parameters(super().sample(sample_shape))

    def _to_parameters(self, noise):
        A_inv = torch.tensor([[1.0, 1, 0], [0, 1, 0], [0, 0, 1]])
        return noise @ A_inv

    def _to_noise(self, parameters):
        A = torch.tensor([[1.0, -1, 0], [0, 1, 0], [0, 0, 1]])
        return parameters @ A


class LotkaVolterraOscillating:
    def __init__(self):
        mean = torch.log(torch.tensor([0.01, 0.5, 1, 0.01]))
        sigma = 0.5
        covariance = sigma ** 2 * torch.eye(4)
        self._gaussian = distributions.MultivariateNormal(
            loc=mean, covariance_matrix=covariance
        )
        self._uniform = BoxUniform(low=-5 * torch.ones(4), high=2 * torch.ones(4))
        self._log_normalizer = -torch.log(
            torch.erf((2 - mean) / sigma) - torch.erf((-5 - mean) / sigma)
        ).sum()

    def log_prob(self, value):
        unnormalized_log_prob = self._gaussian.log_prob(value) + self._uniform.log_prob(
            value
        )

        return self._log_normalizer + unnormalized_log_prob

    def sample(self, sample_shape=torch.Size()):
        num_remaining_samples = sample_shape[0]
        samples = []
        while num_remaining_samples > 0:
            candidate_samples = self._gaussian.sample((num_remaining_samples,))

            uniform_log_prob = self._uniform.log_prob(candidate_samples)

            accepted_samples = candidate_samples[~torch.isinf(uniform_log_prob)]
            samples.append(accepted_samples.detach())

            num_accepted = (~torch.isinf(uniform_log_prob)).sum().item()
            num_remaining_samples -= num_accepted

        # Aggregate collected samples.
        samples = torch.cat(samples)

        # Make sure we have the right amount.
        samples = samples[: sample_shape[0], ...]
        assert samples.shape[0] == sample_shape[0]

        return samples



import functools
from contextlib import contextmanager

import torch.distributed as dist

__all__ = [
    "print0",
    "get_rank",
    "get_world_size",
    "is_rank0",
    "is_dist",
    "rank0_first",
    "rank0_only",
    "get_world_size",
    "barrier",
]


def print0(*args, **kwargs):
    if get_rank() == 0:
        print(*args, **kwargs)


def get_rank():
    return dist.get_rank() if dist.is_initialized() else 0


def get_world_size():
    return dist.get_world_size() if dist.is_initialized() else 1


def is_rank0():
    return not dist.is_initialized() or dist.get_rank() == 0


def is_dist():
    return dist.is_initialized()


def barrier():
    if is_dist():
        dist.barrier()


@contextmanager
def rank0_first():
    if not is_rank0():
        barrier()
    yield
    if dist.is_initialized() and is_rank0():
        barrier()


def rank0_only(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):  # pylint: disable=inconsistent-return-statements
        if is_rank0():
            return func(*args, **kwargs)
        return

    return new_func



from jammy.utils.debug import decorate_exception_hook


# pylint: disable=invalid-name
@decorate_exception_hook
def divide_zero():
    a = 10
    b = a / 0
    return b


if __name__ == "__main__":
    divide_zero()



import matplotlib.pyplot as plt
import numpy as np
import torch as th
from einops import rearrange
from PIL import Image
from torchvision.utils import make_grid

__all__ = [
    "show_batch_img",
    "save_batch_img",
    "show_imgs_traj",
]


def _reshape_viz_batch_img(img_data, shape=7):
    """
    Reshapes the visualization batch image.

    :param img_data: The input image data.
    :param shape: The shape of the output image grid. Can be an integer or a string in the format "x".
    :return: The reshaped image, number of rows, and number of columns.
    :raises RuntimeError: If the shape is not supported.
    """
    if isinstance(shape, int):
        nrow, ncol = shape, shape
    elif isinstance(shape, str):
        if "x" not in shape:
            nrow, ncol = int(shape), int(shape)
        else:
            shape = shape.split("x")
            nrow, ncol = int(shape[0]), int(shape[1])
    else:
        raise RuntimeError(f"shape {shape} not support")
    if isinstance(img_data, th.Tensor):
        assert img_data.shape[1] in [1, 3]
        grid_img = make_grid(img_data[: nrow * ncol].detach().cpu(), ncol)
        img = grid_img.permute(1, 2, 0)
    elif isinstance(img_data, np.ndarray):
        if img_data.shape[1] in [1, 3]:
            img = rearrange(
                img_data[: nrow * ncol], "(b t) c h w -> (b h) (t w) c", b=nrow
            )
        else:
            img = rearrange(
                img_data[: nrow * ncol], "(b t) h w c -> (b h) (t w) c", b=nrow
            )
    return img, nrow, ncol


def show_batch_img(img_data, shape=7, grid=3, is_n1p1=False, auto_n1p1=True):
    """
    Displays a batch of images.

    :param img_data: The batch of images to display.
    :param shape: The shape of the grid to display the images in. Default is 7.
    :param grid: The size of each image in the grid. Default is 3.
    :param is_n1p1: Whether the image data is in the range of -1 to 1. Default is False.
    :param auto_n1p1: Whether to automatically adjust the image data to the range of -1 to 1. Default is True.
    """
    if is_n1p1:
        img_data = (img_data + 1) / 2
    else:
        if auto_n1p1:
            if isinstance(img_data, th.Tensor):
                if img_data.min().item() < -0.5:
                    img_data = (img_data + 1) / 2
            elif isinstance(img_data, np.ndarray):
                if np.min(img_data) < -0.5:
                    img_data = (img_data + 1) / 2
    img, nrow, ncol = _reshape_viz_batch_img(img_data, shape)
    plt.figure(figsize=(ncol * grid, nrow * grid))
    plt.axis("off")
    plt.imshow(img)


def save_batch_img(fpath, img_data, shape=7):
    """
    Saves a batch of images to a file.

    :param fpath: The file path to save the images.
    :param img_data: The batch of images to save.
    :param shape: The shape of the images in the batch (default is 7).
    """
    img, _, _ = _reshape_viz_batch_img(img_data, shape)
    if isinstance(img, np.ndarray):
        img = th.from_numpy(img)
    ndarr = img.mul(255).add_(0.5).clamp_(0, 255).to("cpu", th.uint8).numpy()
    im = Image.fromarray(ndarr)
    im.save(fpath)


def show_imgs_traj(traj_batch_img, num_img, num_steps, grid=3, is_n1p1=False, auto_n1p1=True):
    """
    Displays a trajectory of images.

    :param traj_batch_img: The batch of trajectory images.
    :param num_img: The number of images to display per step.
    :param num_steps: The number of steps in the trajectory.
    :param grid: The number of images to display in each row and column.
    :param is_n1p1: Whether the image values are normalized between -1 and 1.
    :param auto_n1p1: Whether to automatically normalize the image values between -1 and 1.
    """
    idx = np.linspace(0, len(traj_batch_img) - 1, num_steps, dtype=int)
    imgs = []
    for cur_idx in idx:
        imgs.append(traj_batch_img[cur_idx][:num_img])
    imgs = th.cat(imgs)

    imgs = rearrange(imgs, "(n b) ... -> (b n) ...", b=num_img)
    show_batch_img(imgs, f"{num_img}x{num_steps}", grid, is_n1p1, auto_n1p1)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : zmq_utils.py
# Author : Jiayuan Mao, Qsh.Zh
# Email  : qsh.zh27@gmail.com
# Date   : 11/17/2021
#
# Qinsheng modifies based on Jacinle.
# Distributed under terms of the MIT license.

import collections
import contextlib
import queue
import threading

import zmq

from jammy.concurrency.packing import dumpb, loadb
from jammy.concurrency.zmq_utils import bind_to_random_ipc, graceful_close
from jammy.logging import get_logger
from jammy.utils.meta import notnone_property
from jammy.utils.registry import CallbackRegistry

from .utils import get_local_addr

logger = get_logger()

__all__ = ["ServerPipe", "ClientPipe", "make_cs_pair"]

_QueryMessage = collections.namedtuple("QueryMessage", ["identifier", "payload"])


class ServerPipe:  # pylint: disable=too-many-instance-attributes
    def __init__(self, name, send_qsize=0, mode="tcp"):
        self._name = name
        self._conn_info = None

        self._context_lock = threading.Lock()
        self._context = zmq.Context()
        self._tosock = self._context.socket(zmq.ROUTER)
        self._frsock = self._context.socket(zmq.PULL)
        self._tosock.set_hwm(10)
        self._frsock.set_hwm(10)
        self._dispatcher = CallbackRegistry()

        self._send_queue = queue.Queue(maxsize=send_qsize)
        self._rcv_thread = None
        self._snd_thread = None
        self._mode = mode
        assert mode in ("ipc", "tcp")

    @property
    def dispatcher(self):
        return self._dispatcher

    @notnone_property
    def conn_info(self):
        return self._conn_info

    def initialize(self, tcp_port=None):
        self._conn_info = []
        if self._mode == "tcp":
            if tcp_port is not None:
                port = tcp_port[0]
                self._frsock.bind("tcp://*:{}".format(port))
            else:
                port = self._frsock.bind_to_random_port("tcp://*")
            self._conn_info.append("tcp://{}:{}".format(get_local_addr(), port))
            if tcp_port is not None:
                port = tcp_port[1]
                self._tosock.bind("tcp://*:{}".format(port))
            else:
                port = self._tosock.bind_to_random_port("tcp://*")
            self._conn_info.append("tcp://{}:{}".format(get_local_addr(), port))
        elif self._mode == "ipc":
            self._conn_info.append(
                bind_to_random_ipc(self._frsock, self._name + "-c2s-")
            )
            self._conn_info.append(
                bind_to_random_ipc(self._tosock, self._name + "-s2c-")
            )

        self._rcv_thread = threading.Thread(target=self.mainloop_recv, daemon=True)
        self._rcv_thread.start()
        self._snd_thread = threading.Thread(target=self.mainloop_send, daemon=True)
        self._snd_thread.start()

    def finalize(self):
        graceful_close(self._tosock)
        graceful_close(self._frsock)
        self._context.term()

    @contextlib.contextmanager
    def activate(self, tcp_port=None):
        self.initialize(tcp_port=tcp_port)
        try:
            yield
        finally:
            self.finalize()

    def mainloop_recv(self):
        try:
            while True:
                if self._frsock.closed:
                    break

                msg = loadb(self._frsock.recv(copy=False).bytes)
                identifier, data_type, payload = msg
                self._dispatcher.dispatch(data_type, self, identifier, payload)
        except zmq.ContextTerminated:
            pass
        except zmq.ZMQError as e:
            if self._tosock.closed:
                logger.warning("Recv socket closed unexpectedly.")
            else:
                raise e

    def mainloop_send(self):
        try:
            while True:
                if self._tosock.closed:
                    break

                job = self._send_queue.get()
                self._tosock.send_multipart(
                    [job.identifier, dumpb(job.payload)], copy=False
                )
        except zmq.ContextTerminated:
            pass
        except zmq.ZMQError as e:
            if self._tosock.closed:
                logger.warning("Send socket closed unexpectedly.")
            else:
                raise e

    def send(self, identifier, msg):
        self._send_queue.put(_QueryMessage(identifier, msg))


class ClientPipe:
    def __init__(self, name, conn_info):
        self._name = name
        self._conn_info = conn_info
        self._context = None
        self._tosock = None
        self._frsock = None

    @property
    def identity(self):
        return self._name.encode("utf-8")

    def initialize(self):
        self._context = zmq.Context()
        self._tosock = self._context.socket(zmq.PUSH)
        self._frsock = self._context.socket(zmq.DEALER)
        self._tosock.setsockopt(zmq.IDENTITY, self.identity)
        self._frsock.setsockopt(zmq.IDENTITY, self.identity)
        self._tosock.set_hwm(2)
        self._tosock.connect(self._conn_info[0])
        self._frsock.connect(self._conn_info[1])

    def finalize(self):
        graceful_close(self._frsock)
        graceful_close(self._tosock)
        self._context.term()

    @contextlib.contextmanager
    def activate(self):
        self.initialize()
        try:
            yield
        finally:
            self.finalize()

    def query(self, data_type, inp=None, do_recv=True):
        self._tosock.send(dumpb((self.identity, data_type, inp)), copy=False)
        if do_recv:
            return self.recv()
        return None

    def recv(self):
        out = loadb(self._frsock.recv(copy=False).bytes)
        return out


def make_cs_pair(name, nr_clients=None, mode="tcp", send_qsize=10):
    rep = ServerPipe(name + "-rep", mode=mode, send_qsize=send_qsize)
    rep.initialize()
    nr_reqs = nr_clients or 1
    reqs = [ClientPipe(name + "-req-" + str(i), rep.conn_info) for i in range(nr_reqs)]

    if nr_clients is None:
        return rep, reqs[0]
    return rep, reqs



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : matching.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 03/02/2017
#
# Qinsheng modified based on Jacinle.
# Distributed under terms of the MIT license.


"""Functions to match names using glob patterns."""

import collections
import fnmatch
import re
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

__all__ = ["NameMatcher", "IENameMatcher"]


class NameMatcher:
    """A name matcher based on a set of glob patterns.

    The rule set is a list of (pattern, value) pairs. The pattern is a glob pattern,
    and the value is the value to be returned when the pattern matches.

    Example:
        .. code-block:: python

            matcher = NameMatcher({'*.jpg': 'image', '*.png': 'image', '*.txt': 'text'})
            with matcher:
                matcher.match('a.jpg')  # 'image'
                matcher.match('a.png')  # 'image'

            matched, unused = matcher.get_last_stat()  # Return a tuple of (matched values, unmatched patterns).
            print(matched)  # [('a.jpg', '*.jpg', 'image'), ('a.png', '*.png', 'image')]
            print(unused)  # {'*.txt'}
    """

    def __init__(
        self, rules: Optional[Union[Iterable[Tuple[str, Any]], Dict[str, Any]]] = None
    ):
        """Initialize the name matcher.

        Args:
            rules: A list of (pattern, value) pairs, or a dict of {pattern: value}.
        """
        if rules is None:
            self._rules = []
        elif isinstance(rules, dict):
            self._rules = list(rules.items())
        else:
            assert isinstance(rules, collections.Iterable)
            self._rules = list(rules)

        self._map = {}
        self._compiled_rules = []
        self._compiled = False

        self._matched = []
        self._unused = set()
        self._last_stat = None

    @property
    def rules(self) -> List[Tuple[str, Any]]:
        """Get the rules."""
        return self._rules

    def map(self) -> Dict[str, Any]:
        """Get the map of {pattern: value}."""
        assert self._compiled
        return self._map

    def append_rule(self, rule: Tuple[str, Any]):
        """Append a rule to the rule set. The rule is a (pattern, value) pair.

        Args:
            rule: the rule to be appended.
        """
        self._rules.append(tuple(rule))

    def insert_rule(self, index: int, rule: Tuple[str, Any]):
        """Insert a rule to the rule set at a given position (priority). The rule is a (pattern, value) pair."""
        self._rules.insert(index, rule)

    def pop_rule(self, index=None):
        """Pop a rule from the rule set.

        Args:
            index: the index of the rule to be popped. If None, the last rule will be popped.
        """
        self._rules.pop(index)

    def begin(self, *, force_compile=False):
        """Begin a matching session."""
        if not self._compiled or force_compile:
            self.compile()
        self._matched = []
        self._unused = set(range(len(self._compiled_rules)))

    def end(self):
        """End a matching session, which returns a tuple of (matched values, unmatched patterns).
        See the docstring of :class:`NameMatcher` for more details."""
        return self._matched, {self._compiled_rules[i][0] for i in self._unused}

    def match(self, k: str) -> Optional[Any]:
        """Match a name against the rule set. Return the value if matched, otherwise return None.

        Args:
            k: the name to be matched.

        Returns:
            The value if matched, otherwise None.
        """
        for i, (r, p, v) in enumerate(self._compiled_rules):
            if p.match(k):
                if i in self._unused:
                    self._unused.remove(i)
                self._matched.append((k, r, v))
                return v
        return None

    def compile(self):
        """Compile the rule set."""
        self._map = {}
        self._compiled_rules = []

        for r, v in self._rules:
            self._map[r] = v
            p = fnmatch.translate(r)
            p = re.compile(p, flags=re.IGNORECASE)
            self._compiled_rules.append((r, p, v))
        self._compiled = True

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._last_stat = self.end()

    def get_last_stat(self):
        """Get the last matching session's result."""
        return self._last_stat


class IENameMatcher:
    """A name matcher based on two sets of glob patterns: one for inclusion and one for exclusion.

    - When ``include`` is None, ``exclude`` is not None, the matcher will match all names that are not excluded.
    - When ``include`` is not None, ``exclude`` is None, the matcher will match all names that are included.
    - When ``include`` is not None, ``exclude`` is not None,
        the matcher will match all names that are included and not excluded.
        The ``exclude`` rule set has higher priority than the ``include`` rule set.

    Example:
        .. code-block:: python

            matcher = IENameMatcher(include=['*.jpg', '*.png'], exclude=['*.bak.png'])
            with matcher:
                matcher.match('a.jpg')  # True
                matcher.match('a.png')  # True
                matcher.match('a.bak.png')  # False
                matcher.match('a.txt')  # False
                matcher.match('a.bak.txt')  # False

            stat_type, things = matcher.get_last_stat()
            print(stat_type)  # 'exclude'
            # Everything that has been rejected.
            print(things)  # ['a.bak.png', 'a.txt', 'a.bak.txt']
    """

    def __init__(
        self, include: Optional[Iterable[str]], exclude: Optional[Iterable[str]]
    ):
        """Initialize the name matcher.

        Args:
            include: a list of glob patterns for inclusion.
            exclude: a list of glob patterns for exclusion.
        """
        if include is None:
            self.include = None
        else:
            self.include = NameMatcher([(i, True) for i in include])

        if exclude is None:
            self.exclude = None
        else:
            self.exclude = NameMatcher([(e, True) for e in exclude])
        self._last_stat = None

    def begin(self):
        """Begin a matching session."""
        if self.include is not None:
            self.include.begin()
        if self.exclude is not None:
            self.exclude.begin()
        self._last_stat = (set(), set())

    def end(self):
        """End a matching session, which returns a tuple of ``(stat_type, things)``
        See the docstring of :class:`IENameMatcher`."""
        if self.include is not None:
            self.include.end()
        if self.exclude is not None:
            self.exclude.end()

        if len(self._last_stat[0]) < len(self._last_stat[1]):
            self._last_stat = ("included", self._last_stat[0])
        else:
            self._last_stat = ("excluded", self._last_stat[1])

    def match(self, k: str) -> bool:
        """Match a name against the rule set. Return True if matched, otherwise return False."""
        if self.include is None:
            ret = True
        else:
            ret = bool(self.include.match(k))

        if self.exclude is not None:
            ret = ret and not bool(self.exclude.match(k))

        if ret:
            self._last_stat[0].add(k)
        else:
            self._last_stat[1].add(k)
        return ret

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end()

    def get_last_stat(self):
        """Get the last matching session's result."""
        return self._last_stat



# Copyright (c) OpenMMLab. All rights reserved.
from .base import BaseFileHandler
from .json_handler import JsonHandler
from .pickle_handler import PickleHandler
from .registry_utils import file_handlers, register_handler
from .yaml_handler import YamlHandler

__all__ = [
    "BaseFileHandler",
    "JsonHandler",
    "PickleHandler",
    "YamlHandler",
    "register_handler",
    "file_handlers",
]



from __future__ import annotations

import math
import os.path as osp

__all__ = ["fsize_format", "get_ext", "get_name"]

unit_list = list(zip(["bytes", "kB", "MB", "GB", "TB", "PB"], [0, 0, 1, 2, 2, 2]))


def get_ext(fname, match_first=False):
    if match_first:
        fname = osp.split(fname)[1]
        return fname[fname.find(".") :]
    else:
        return osp.splitext(fname)[1]


def get_name(fname, match_first=False):
    ext = get_ext(fname, match_first)
    return fname[: -len(ext)]


def fsize_format(num):
    """Human readable file size."""
    # from http://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size

    if num == 0:
        return "0 bytes"
    if num == 1:
        return "1 byte"

    exponent = min(int(math.log(num, 1024)), len(unit_list) - 1)
    quotient = float(num) / 1024**exponent
    unit, num_decimals = unit_list[exponent]
    format_string = f"{{:.{num_decimals}f}} {{}}"
    return format_string.format(quotient, unit)



# Copyright (c) OpenMMLab. All rights reserved.
import io
import os
import os.path as osp
import re
import tempfile
from contextlib import contextmanager
from pathlib import Path
from shutil import SameFileError
from typing import Generator, Iterator, Optional, Tuple, Union

from jammy.io.path import mkdir_or_exist
from jammy.utils import has_method

from .base import BaseStorageBackend
from .boto3_client import Boto3Client


class Boto3Backend(BaseStorageBackend):
    """boto3 storage backend (for internal usage).

    Boto3Backend supports reading and writing data to multiple clusters.
    If the file path contains the cluster name, Boto3Backend will read data
    from specified cluster or write data to it. Otherwise, Boto3Backend will
    access the default cluster.

    Args:
        path_mapping (dict, optional): Path mapping dict from local path to
            Boto3 path. When ``path_mapping={'src': 'dst'}``, ``src`` in
            ``filepath`` will be replaced by ``dst``. Defaults to None.
        enable_mc (bool, optional): Whether to enable memcached support.
            Defaults to True.
        conf_path (str, optional): Config path of Boto3 client. Default: None.
            `New in version 0.3.3`.

    Examples:
        >>> backend = Boto3Backend()
        >>> filepath1 = 's3://path/of/file'
        >>> filepath2 = 'cluster-name:s3://path/of/file'
        >>> backend.get(filepath1)  # get data from default cluster
        >>> client.get(filepath2)  # get data from 'cluster-name' cluster
    """

    def __init__(
        self,
        path_mapping: Optional[dict] = None,
        enable_mc: bool = True,
        conf_path: Optional[str] = None,
    ):
        self._client = Boto3Client(conf_path=conf_path, enable_mc=enable_mc)
        assert isinstance(path_mapping, dict) or path_mapping is None
        self.path_mapping = path_mapping

    def _map_path(self, filepath: Union[str, Path]) -> str:
        """Map ``filepath`` to a string path whose prefix will be replaced by
        :attr:`self.path_mapping`.

        Args:
            filepath (str or Path): Path to be mapped.
        """
        filepath = str(filepath)
        if self.path_mapping is not None:
            for k, v in self.path_mapping.items():
                filepath = filepath.replace(k, v, 1)
        return filepath

    def _format_path(self, filepath: str) -> str:
        """Convert a ``filepath`` to standard format of s3 oss.

        If the ``filepath`` is concatenated by ``os.path.join``, in a Windows
        environment, the ``filepath`` will be the format of
        's3://bucket_name\\image.jpg'. By invoking :meth:`_format_path`, the
        above ``filepath`` will be converted to 's3://bucket_name/image.jpg'.

        Args:
            filepath (str): Path to be formatted.
        """
        return re.sub(r"\\+", "/", filepath)

    def _replace_prefix(self, filepath: Union[str, Path]) -> str:
        filepath = str(filepath)
        return filepath
        # return filepath.replace('s3://', 's3://')

    def get(self, filepath: Union[str, Path]) -> bytes:
        """Read bytes from a given ``filepath`` with 'rb' mode.

        Args:
            filepath (str or Path): Path to read data.

        Returns:
            bytes: Return bytes read from filepath.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/file'
            >>> backend.get(filepath)
            b'hello world'
        """
        filepath = self._map_path(filepath)
        filepath = self._format_path(filepath)
        filepath = self._replace_prefix(filepath)
        value = self._client.get(filepath)
        return value

    def get_text(
        self,
        filepath: Union[str, Path],
        encoding: str = "utf-8",
    ) -> str:
        """Read text from a given ``filepath`` with 'r' mode.

        Args:
            filepath (str or Path): Path to read data.
            encoding (str): The encoding format used to open the ``filepath``.
                Defaults to 'utf-8'.

        Returns:
            str: Expected text reading from ``filepath``.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/file'
            >>> backend.get_text(filepath)
            'hello world'
        """
        return str(self.get(filepath), encoding=encoding)

    def put(self, obj: Union[bytes, io.BytesIO], filepath: Union[str, Path]) -> None:
        """Write bytes to a given ``filepath``.

        Args:
            obj (bytes): Data to be saved.
            filepath (str or Path): Path to write data.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/file'
            >>> backend.put(b'hello world', filepath)
        """
        filepath = self._map_path(filepath)
        filepath = self._format_path(filepath)
        filepath = self._replace_prefix(filepath)
        self._client.put(obj, filepath)

    def put_text(
        self,
        obj: str,
        filepath: Union[str, Path],
        encoding: str = "utf-8",
    ) -> None:
        """Write text to a given ``filepath``.

        Args:
            obj (str): Data to be written.
            filepath (str or Path): Path to write data.
            encoding (str): The encoding format used to encode the ``obj``.
                Defaults to 'utf-8'.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/file'
            >>> backend.put_text('hello world', filepath)
        """
        self.put(bytes(obj, encoding=encoding), filepath)

    def exists(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path exists.

        Args:
            filepath (str or Path): Path to be checked whether exists.

        Returns:
            bool: Return ``True`` if ``filepath`` exists, ``False`` otherwise.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/file'
            >>> backend.exists(filepath)
            True
        """
        if not (
            has_method(self._client, "contains") and has_method(self._client, "isdir")
        ):
            raise NotImplementedError(
                "Current version of Boto3 Python SDK has not supported "
                "the `contains` and `isdir` methods, please use a higher"
                "version or dev branch instead."
            )

        filepath = self._map_path(filepath)
        filepath = self._format_path(filepath)
        filepath = self._replace_prefix(filepath)
        return self._client.contains(filepath) or self._client.isdir(filepath)

    def isdir(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a directory.

        Args:
            filepath (str or Path): Path to be checked whether it is a
                directory.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a directory,
            ``False`` otherwise.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/dir'
            >>> backend.isdir(filepath)
            True
        """
        if not has_method(self._client, "isdir"):
            raise NotImplementedError(
                "Current version of Boto3 Python SDK has not supported "
                "the `isdir` method, please use a higher version or dev"
                " branch instead."
            )

        filepath = self._map_path(filepath)
        filepath = self._format_path(filepath)
        filepath = self._replace_prefix(filepath)
        return self._client.isdir(filepath)

    def isfile(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a file.

        Args:
            filepath (str or Path): Path to be checked whether it is a file.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a file, ``False``
            otherwise.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/file'
            >>> backend.isfile(filepath)
            True
        """
        if not has_method(self._client, "contains"):
            raise NotImplementedError(
                "Current version of Boto3 Python SDK has not supported "
                "the `contains` method, please use a higher version or "
                "dev branch instead."
            )

        filepath = self._map_path(filepath)
        filepath = self._format_path(filepath)
        filepath = self._replace_prefix(filepath)
        return self._client.contains(filepath)

    def join_path(
        self,
        filepath: Union[str, Path],
        *filepaths: Union[str, Path],
    ) -> str:
        r"""Concatenate all file paths.

        Join one or more filepath components intelligently. The return value
        is the concatenation of filepath and any members of \*filepaths.

        Args:
            filepath (str or Path): Path to be concatenated.

        Returns:
            str: The result after concatenation.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/file'
            >>> backend.join_path(filepath, 'another/path')
            's3://path/of/file/another/path'
            >>> backend.join_path(filepath, '/another/path')
            's3://path/of/file/another/path'
        """
        filepath = self._format_path(self._map_path(filepath))
        if filepath.endswith("/"):
            filepath = filepath[:-1]
        formatted_paths = [filepath]
        for path in filepaths:
            formatted_path = self._format_path(self._map_path(path))
            formatted_paths.append(formatted_path.lstrip("/"))

        return "/".join(formatted_paths)

    @contextmanager
    def get_local_path(
        self,
        filepath: Union[str, Path],
    ) -> Generator[Union[str, Path], None, None]:
        """Download a file from ``filepath`` to a local temporary directory,
        and return the temporary path.

        ``get_local_path`` is decorated by :meth:`contxtlib.contextmanager`. It
        can be called with ``with`` statement, and when exists from the
        ``with`` statement, the temporary path will be released.

        Args:
            filepath (str or Path): Download a file from ``filepath``.

        Yields:
            Iterable[str]: Only yield one temporary path.

        Examples:
            >>> backend = Boto3Backend()
            >>> # After existing from the ``with`` clause,
            >>> # the path will be removed
            >>> filepath = 's3://path/of/file'
            >>> with backend.get_local_path(filepath) as path:
            ...     # do something here
        """
        assert self.isfile(filepath)
        try:
            f = tempfile.NamedTemporaryFile(delete=False)
            f.write(self.get(filepath))
            f.close()
            yield f.name
        finally:
            os.remove(f.name)

    def copyfile(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Copy a file src to dst and return the destination file.

        src and dst should have the same prefix. If dst specifies a directory,
        the file will be copied into dst using the base filename from src. If
        dst specifies a file that already exists, it will be replaced.

        Args:
            src (str or Path): A file to be copied.
            dst (str or Path): Copy file to dst.

        Returns:
            str: The destination file.

        Raises:
            SameFileError: If src and dst are the same file, a SameFileError
                will be raised.

        Examples:
            >>> backend = Boto3Backend()
            >>> # dst is a file
            >>> src = 's3://path/of/file'
            >>> dst = 's3://path/of/file1'
            >>> backend.copyfile(src, dst)
            's3://path/of/file1'

            >>> # dst is a directory
            >>> dst = 's3://path/of/dir'
            >>> backend.copyfile(src, dst)
            's3://path/of/dir/file'
        """
        src = self._format_path(self._map_path(src))
        dst = self._format_path(self._map_path(dst))
        if self.isdir(dst):
            dst = self.join_path(dst, src.split("/")[-1])

        if src == dst:
            raise SameFileError("src and dst should not be same")

        self.put(self.get(src), dst)
        return dst

    def copytree(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Recursively copy an entire directory tree rooted at src to a
        directory named dst and return the destination directory.

        src and dst should have the same prefix.

        Args:
            src (str or Path): A directory to be copied.
            dst (str or Path): Copy directory to dst.
            backend_args (dict, optional): Arguments to instantiate the
                prefix of uri corresponding backend. Defaults to None.

        Returns:
            str: The destination directory.

        Raises:
            FileExistsError: If dst had already existed, a FileExistsError will
                be raised.

        Examples:
            >>> backend = Boto3Backend()
            >>> src = 's3://path/of/dir'
            >>> dst = 's3://path/of/dir1'
            >>> backend.copytree(src, dst)
            's3://path/of/dir1'
        """
        src = self._format_path(self._map_path(src))
        dst = self._format_path(self._map_path(dst))

        if self.exists(dst):
            raise FileExistsError("dst should not exist")

        for path in self.list_dir_or_file(src, list_dir=False, recursive=True):
            src_path = self.join_path(src, path)
            dst_path = self.join_path(dst, path)
            self.put(self.get(src_path), dst_path)

        return dst

    def copyfile_from_local(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Upload a local file src to dst and return the destination file.

        Args:
            src (str or Path): A local file to be copied.
            dst (str or Path): Copy file to dst.
            backend_args (dict, optional): Arguments to instantiate the
                prefix of uri corresponding backend. Defaults to None.

        Returns:
            str: If dst specifies a directory, the file will be copied into dst
            using the base filename from src.

        Examples:
            >>> backend = Boto3Backend()
            >>> # dst is a file
            >>> src = 'path/of/your/file'
            >>> dst = 's3://path/of/file1'
            >>> backend.copyfile_from_local(src, dst)
            's3://path/of/file1'

            >>> # dst is a directory
            >>> dst = 's3://path/of/dir'
            >>> backend.copyfile_from_local(src, dst)
            's3://path/of/dir/file'
        """
        dst = self._format_path(self._map_path(dst))
        if self.isdir(dst):
            dst = self.join_path(dst, osp.basename(src))

        with open(src, "rb") as f:
            self.put(f.read(), dst)

        return dst

    def copytree_from_local(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Recursively copy an entire directory tree rooted at src to a
        directory named dst and return the destination directory.

        Args:
            src (str or Path): A local directory to be copied.
            dst (str or Path): Copy directory to dst.

        Returns:
            str: The destination directory.

        Raises:
            FileExistsError: If dst had already existed, a FileExistsError will
                be raised.

        Examples:
            >>> backend = Boto3Backend()
            >>> src = 'path/of/your/dir'
            >>> dst = 's3://path/of/dir1'
            >>> backend.copytree_from_local(src, dst)
            's3://path/of/dir1'
        """
        dst = self._format_path(self._map_path(dst))
        if self.exists(dst):
            raise FileExistsError("dst should not exist")

        src = str(src)

        for cur_dir, _, files in os.walk(src):
            for f in files:
                src_path = osp.join(cur_dir, f)
                dst_path = self.join_path(dst, src_path.replace(src, ""))
                self.copyfile_from_local(src_path, dst_path)

        return dst

    def copyfile_to_local(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> Union[str, Path]:
        """Copy the file src to local dst and return the destination file.

        If dst specifies a directory, the file will be copied into dst using
        the base filename from src. If dst specifies a file that already
        exists, it will be replaced.

        Args:
            src (str or Path): A file to be copied.
            dst (str or Path): Copy file to to local dst.

        Returns:
            str: If dst specifies a directory, the file will be copied into dst
            using the base filename from src.

        Examples:
            >>> backend = Boto3Backend()
            >>> # dst is a file
            >>> src = 's3://path/of/file'
            >>> dst = 'path/of/your/file'
            >>> backend.copyfile_to_local(src, dst)
            'path/of/your/file'

            >>> # dst is a directory
            >>> dst = 'path/of/your/dir'
            >>> backend.copyfile_to_local(src, dst)
            'path/of/your/dir/file'
        """
        if osp.isdir(dst):
            basename = osp.basename(src)
            if isinstance(dst, str):
                dst = osp.join(dst, basename)
            else:
                assert isinstance(dst, Path)
                dst = dst / basename

        with open(dst, "wb") as f:
            f.write(self.get(src))

        return dst

    def copytree_to_local(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> Union[str, Path]:
        """Recursively copy an entire directory tree rooted at src to a local
        directory named dst and return the destination directory.

        Args:
            src (str or Path): A directory to be copied.
            dst (str or Path): Copy directory to local dst.
            backend_args (dict, optional): Arguments to instantiate the
                prefix of uri corresponding backend. Defaults to None.

        Returns:
            str: The destination directory.

        Examples:
            >>> backend = Boto3Backend()
            >>> src = 's3://path/of/dir'
            >>> dst = 'path/of/your/dir'
            >>> backend.copytree_to_local(src, dst)
            'path/of/your/dir'
        """
        for path in self.list_dir_or_file(src, list_dir=False, recursive=True):
            dst_path = osp.join(dst, path)
            mkdir_or_exist(osp.dirname(dst_path))
            with open(dst_path, "wb") as f:
                f.write(self.get(self.join_path(src, path)))

        return dst

    def remove(self, filepath: Union[str, Path]) -> None:
        """Remove a file.

        Args:
            filepath (str or Path): Path to be removed.

        Raises:
            FileNotFoundError: If filepath does not exist, an FileNotFoundError
                will be raised.
            IsADirectoryError: If filepath is a directory, an IsADirectoryError
                will be raised.

        Examples:
            >>> backend = Boto3Backend()
            >>> filepath = 's3://path/of/file'
            >>> backend.remove(filepath)
        """
        if not has_method(self._client, "delete"):
            raise NotImplementedError(
                "Current version of Boto3 Python SDK has not supported "
                "the `delete` method, please use a higher version or dev "
                "branch instead."
            )

        if not self.exists(filepath):
            raise FileNotFoundError(f"filepath {filepath} does not exist")

        if self.isdir(filepath):
            raise IsADirectoryError("filepath should be a file")

        filepath = self._map_path(filepath)
        filepath = self._format_path(filepath)
        filepath = self._replace_prefix(filepath)
        self._client.delete(filepath)

    def rmtree(self, dir_path: Union[str, Path]) -> None:
        """Recursively delete a directory tree.

        Args:
            dir_path (str or Path): A directory to be removed.

        Examples:
            >>> backend = Boto3Backend()
            >>> dir_path = 's3://path/of/dir'
            >>> backend.rmtree(dir_path)
        """
        for path in self.list_dir_or_file(dir_path, list_dir=False, recursive=True):
            filepath = self.join_path(dir_path, path)
            self.remove(filepath)

    def copy_if_symlink_fails(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> bool:
        """Create a symbolic link pointing to src named dst.

        Directly copy src to dst because PetrelBacekend does not support create
        a symbolic link.

        Args:
            src (str or Path): A file or directory to be copied.
            dst (str or Path): Copy a file or directory to dst.
            backend_args (dict, optional): Arguments to instantiate the
                prefix of uri corresponding backend. Defaults to None.

        Returns:
            bool: Return False because Boto3Backend does not support create
            a symbolic link.

        Examples:
            >>> backend = Boto3Backend()
            >>> src = 's3://path/of/file'
            >>> dst = 's3://path/of/your/file'
            >>> backend.copy_if_symlink_fails(src, dst)
            False
            >>> src = 's3://path/of/dir'
            >>> dst = 's3://path/of/your/dir'
            >>> backend.copy_if_symlink_fails(src, dst)
            False
        """
        if self.isfile(src):
            self.copyfile(src, dst)
        else:
            self.copytree(src, dst)
        return False

    def list_dir_or_file(  # pylint: disable=too-many-arguments
        self,
        dir_path: Union[str, Path],
        list_dir: bool = True,
        list_file: bool = True,
        suffix: Optional[Union[str, Tuple[str]]] = None,
        recursive: bool = False,
    ) -> Iterator[str]:
        """Scan a directory to find the interested directories or files in
        arbitrary order.

        Note:
            Boto3 has no concept of directories but it simulates the directory
            hierarchy in the filesystem through public prefixes. In addition,
            if the returned path ends with '/', it means the path is a public
            prefix which is a logical directory.

        Note:
            :meth:`list_dir_or_file` returns the path relative to ``dir_path``.
            In addition, the returned path of directory will not contains the
            suffix '/' which is consistent with other backends.

        Args:
            dir_path (str | Path): Path of the directory.
            list_dir (bool): List the directories. Defaults to True.
            list_file (bool): List the path of files. Defaults to True.
            suffix (str or tuple[str], optional):  File suffix
                that we are interested in. Defaults to None.
            recursive (bool): If set to True, recursively scan the
                directory. Defaults to False.

        Yields:
            Iterable[str]: A relative path to ``dir_path``.

        Examples:
            >>> backend = Boto3Backend()
            >>> dir_path = 's3://path/of/dir'
            >>> # list those files and directories in current directory
            >>> for file_path in backend.list_dir_or_file(dir_path):
            ...     print(file_path)
            >>> # only list files
            >>> for file_path in backend.list_dir_or_file(dir_path, list_dir=False):
            ...     print(file_path)
            >>> # only list directories
            >>> for file_path in backend.list_dir_or_file(dir_path, list_file=False):
            ...     print(file_path)
            >>> # only list files ending with specified suffixes
            >>> for file_path in backend.list_dir_or_file(dir_path, suffix='.txt'):
            ...     print(file_path)
            >>> # list all files and directory recursively
            >>> for file_path in backend.list_dir_or_file(dir_path, recursive=True):
            ...     print(file_path)
        """  # noqa: E501
        if not has_method(self._client, "list"):
            raise NotImplementedError(
                "Current version of Boto3 Python SDK has not supported "
                "the `list` method, please use a higher version or dev"
                " branch instead."
            )

        dir_path = self._map_path(dir_path)
        dir_path = self._format_path(dir_path)
        dir_path = self._replace_prefix(dir_path)
        if list_dir and suffix is not None:
            raise TypeError("`list_dir` should be False when `suffix` is not None")

        if (suffix is not None) and not isinstance(suffix, (str, tuple)):
            raise TypeError("`suffix` must be a string or tuple of strings")

        # Boto3's simulated directory hierarchy assumes that directory paths
        # should end with `/`
        if not dir_path.endswith("/"):
            dir_path += "/"

        root = dir_path

        def _list_dir_or_file(dir_path, list_dir, list_file, suffix, recursive):
            for path in self._client.list(dir_path):
                # the `self.isdir` is not used here to determine whether path
                # is a directory, because `self.isdir` relies on
                # `self._client.list`
                if path.endswith("/"):  # a directory path
                    next_dir_path = self.join_path(dir_path, path)
                    if list_dir:
                        # get the relative path and exclude the last
                        # character '/'
                        rel_dir = next_dir_path[len(root) : -1]
                        yield rel_dir
                    if recursive:
                        yield from _list_dir_or_file(
                            next_dir_path, list_dir, list_file, suffix, recursive
                        )
                else:  # a file path
                    absolute_path = self.join_path(dir_path, path)
                    rel_path = absolute_path[len(root) :]
                    if (suffix is None or rel_path.endswith(suffix)) and list_file:
                        yield rel_path

        return _list_dir_or_file(dir_path, list_dir, list_file, suffix, recursive)

    def generate_presigned_url(
        self, url: str, client_method: str = "get_object", expires_in: int = 3600
    ) -> str:
        """Generate the presigned url of video stream which can be passed to
        mmcv.VideoReader. Now only work on Boto3 backend.

        Note:
            Now only work on Boto3 backend.

        Args:
            url (str): Url of video stream.
            client_method (str): Method of client, 'get_object' or
                'put_object'. Default: 'get_object'.
            expires_in (int): expires, in seconds. Default: 3600.

        Returns:
            str: Generated presigned url.
        """
        raise NotImplementedError(
            "generate_presigned_url is not supported in Boto3Backend"
        )
        return self._client.generate_presigned_url(url, client_method, expires_in)



import os

import hydra

from jammy.cli_scripts.gpu_sc import get_gpu_by_utils


@hydra.main(config_path=".", config_name="config")
def my_app(cfg):
    del cfg
    pid = os.getpid()
    gpus = get_gpu_by_utils(sleep_sec=5)
    print(f"{pid}\t get {gpus}")


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



import threading
import time
from subprocess import PIPE, Popen

import psutil

from jammy.logging import get_logger

__all__ = ["run_simple_command", "BashRunner"]


def run_simple_command(cmd: str):
    cmd = cmd.split(" ")
    with Popen(cmd, stdout=PIPE, stderr=PIPE) as prc:
        stdout, stderr = prc.communicate()
        stdout = stdout.decode("utf-8")
        stderr = stderr.decode("utf-8")
        prc.terminate()
    return stdout, stderr


class ThreadSafeDict(dict):
    def __init__(self, *p_arg, **n_arg):
        dict.__init__(self, *p_arg, **n_arg)
        self._lock = threading.Lock()

    def __enter__(self):
        self._lock.acquire()
        return self

    def __exit__(self, type_p, value, traceback):
        del type_p, value, traceback
        self._lock.release()


def _get_mem():
    return psutil.virtual_memory().available / 1024 / 1024 / 1024


logger = get_logger()


class BashRunner:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        notifier=None,
        interval_cpu_check: float = 1.0,
        cpu_usage_thres: float = 92,
        seconds_numthread_wait: int = 10,
        mem_avail_thres: float = 0.9,
        seconds_resouece_wait: int = 2,
    ):
        self.notifier = notifier
        self.interval_cpu_check = interval_cpu_check
        self.cpu_usage_thres = cpu_usage_thres
        self.mem_avail_thres = mem_avail_thres
        self.seconds_numthread_wait = seconds_numthread_wait
        self.seconds_resouece_wait = seconds_resouece_wait

        self.complete_record = []
        self.active_task = ThreadSafeDict()
        self.error_task = {}
        self.num_total_task = 0

    def proc_event(self, cur_process):
        finished_num = len(self.complete_record) + len(self.error_task)
        return f"\nPID:{cur_process.pid}. Active:{len(self.active_task)}.\
             Finished: {finished_num}/{self.num_total_task}"

    def _execute_single_task(self, str_cmd):
        logger.debug("Checking states")
        while not (self._check_cpu() and self._check_mem()):
            time.sleep(self.seconds_resouece_wait)
        with Popen(str_cmd, shell=True) as process:  # , stdout=PIPE, stderr=PIPE)
            return process

    def _check_single_task(self, cmd, task_process):
        poll_state = task_process.poll()
        if poll_state is None:
            return False
        elif poll_state == 0:
            if cmd not in self.complete_record:
                self._print_finish(cmd, task_process)
        else:
            if cmd not in self.error_task:
                self._print_error(cmd, task_process)
        return True

    def reset(self, batch_cmd):
        self.num_total_task = len(batch_cmd)
        self.complete_record.clear()
        self.active_task.clear()
        self.error_task.clear()

    def execute_tasks(self, batch_cmd, num_thread=0, msg=None):
        self.reset(batch_cmd)
        if num_thread == 1:
            self.queue_tasks(batch_cmd, msg)
        else:
            thread_check = threading.Thread(target=self._check_process)
            thread_check.start()
            for task in batch_cmd:
                self.check_num_thread(num_thread)
                process = self._execute_single_task(task)
                self._print_start(task, process)
                with self.active_task as active_task:
                    active_task[task] = process
            thread_check.join()
            self._send_msg(msg)

    def queue_tasks(self, batch_cmd, msg=None):
        for task in batch_cmd:
            with Popen(task, shell=True) as process:
                self._print_start(task, process)
                process.communicate()
                if process.poll() == 0:
                    self._print_finish(task, process)
                else:
                    self._print_error(task, process)
        self._print_complete()
        self._send_msg(msg)

    def _send_msg(self, msg=None):
        if self.notifier is None:
            return
        if msg == 0:
            return
        else:
            if len(self.error_task) == 0:
                msg += "\n All tasks success!"
            else:
                for cmd, error_code in self.error_task.items():
                    msg += f"\nError: {error_code:03d}\t{cmd}"
            self.notifier.notify(msg)

    def check_num_thread(self, num_thread):
        if num_thread > 0:
            while len(self.active_task) >= num_thread:
                time.sleep(self.seconds_numthread_wait)

    def kill_all(self):
        for cmd, process in self.active_task.items():
            process.kill()
            logger.info(f"Kill {cmd}")

    def _check_process(self):
        logger.debug("Start monitor thread")
        is_exit = False
        while not is_exit:
            delete_active = []
            with self.active_task as active_task:
                for cmd, task_process in active_task.items():
                    if self._check_single_task(cmd, task_process):
                        delete_active.append(cmd)

            for key in delete_active:
                del self.active_task[key]

            time.sleep(5)
            if len(self.complete_record) + len(self.error_task) == self.num_total_task:
                self._print_complete()
                is_exit = True
        logger.debug("Exit check thread")

    def _check_cpu(self):
        logger.debug(f"cpu {psutil.cpu_percent(interval=self.interval_cpu_check)}")
        return self.cpu_usage_thres > psutil.cpu_percent(
            interval=self.interval_cpu_check
        )

    def _check_mem(self):
        logger.debug(f"mem {_get_mem()}")
        return self.mem_avail_thres < _get_mem()

    def _print_finish(self, cmd, task_process):
        self.complete_record.append(cmd)
        logger.info(f"Finish {cmd}! {self.proc_event(task_process)}")

    def _print_error(self, cmd, task_process):
        poll_state = task_process.poll()
        self.error_task[cmd] = poll_state
        logger.critical(
            f"Error:\t{poll_state:03d}! {self.proc_event(task_process)}\n{cmd}"
        )

    def _print_start(self, cmd, task_process):
        logger.info(f"Fire {cmd}.{self.proc_event(task_process)}")

    def _print_complete(self):
        logger.success(
            f"Complete tasks: {len(self.complete_record)}\tError tasks: {len(self.error_task)}"
        )



from __future__ import annotations

import collections
import contextlib
import io
import os
import sys
import threading
from dataclasses import asdict, is_dataclass
from typing import Any, ClassVar, Optional

import numpy as np

from .registry import LockRegistry

__all__ = [
    "PrintToStringContext",
    "SupressPrintCTX",
    "bubbletext",
    "indent_text",
    "kvformat",
    "kvprint",
    "print2format",
    "print_to",
    "print_to_string",
    "stformat",
    "stprint",
]

_DEFAULT_FLOAT_FORMAT = "{:.6f}"


def indent_text(text: str, level: int = 1, indent_format: Optional[str] = None, tabsize: Optional[int] = None) -> str:
    # Replace asserts with ValueError
    if indent_format is not None and tabsize is not None:
        raise ValueError("Cannot provide both indent format and tabsize.")

    if tabsize is not None:
        indent_format = " " * tabsize
    elif indent_format is None:
        indent_format = "  "

    final_indent = indent_format * level
    return final_indent + text.replace("\n", "\n" + final_indent)


def format_printable_data(data: Any, float_format: str = _DEFAULT_FLOAT_FORMAT) -> str:
    t = type(data)
    if t is np.ndarray:
        return f"ndarray{data.shape}, dtype={data.dtype}"
    # Handle torch.tensor
    if "Tensor" in str(t):
        return f"tensor{tuple(data.shape)}, dtype={data.dtype}"
    elif t is float:
        return float_format.format(data)
    else:
        return str(data)


def _process_data_recursive(  # noqa: C901
    data: Any,
    indent: int,
    key: Optional[str],
    max_depth: int,
    indent_format: str,
    end_format: str,
    float_format: str,
    file: Optional[io.TextIOBase],
) -> None:
    """Helper function to process data recursively for stprint."""
    if file is None:
        file = sys.stdout

    def _indent_print(msg: str, indent: int, prefix: Optional[str] = None) -> None:
        print(indent_format * indent, end="", file=file)
        if prefix is not None:
            print(prefix, end="", file=file)
        print(msg, end=end_format, file=file)

    t = type(data)

    if max_depth == 0:
        if t in (tuple, list):
            _indent_print(f"({t.__name__} of length {len(data)}) ...", indent, prefix=key)
        elif is_dataclass(data):
            _indent_print(f"(dataclass {data.__class__.__name__}) ...", indent, prefix=key)
        elif t in (dict, collections.OrderedDict, collections.defaultdict):
            _indent_print(f"(dict of length {len(data)}) ...", indent, prefix=key)
        return

    if t is tuple:
        _indent_print("tuple[", indent, prefix=key)
        for v in data:
            _process_data_recursive(v, indent + 1, None, max_depth - 1, indent_format, end_format, float_format, file)
        _indent_print("]", indent)
    elif t is list:
        _indent_print("list[", indent, prefix=key)
        for v in data:
            _process_data_recursive(v, indent + 1, None, max_depth - 1, indent_format, end_format, float_format, file)
        _indent_print("]", indent)
    elif is_dataclass(data):
        typename = f"dataclass {data.__class__.__name__}"
        data_dict = asdict(data)
        _indent_print(typename + "{", indent, prefix=key)
        for k in sorted(data_dict.keys()):
            v = data_dict[k]
            _process_data_recursive(
                v, indent + 1, f"{k}: ", max_depth - 1, indent_format, end_format, float_format, file
            )
        _indent_print("}", indent)
    elif t in (dict, collections.OrderedDict, collections.defaultdict):
        typename = {
            dict: "dict",
            collections.OrderedDict: "ordered_dict",
            collections.defaultdict: "default_dict",
        }[t]
        _indent_print(typename + "{", indent, prefix=key)
        for k in sorted(data.keys()):
            v = data[k]
            _process_data_recursive(
                v, indent + 1, f"{k}: ", max_depth - 1, indent_format, end_format, float_format, file
            )
        _indent_print("}", indent)
    else:
        _indent_print(format_printable_data(data, float_format=float_format), indent, prefix=key)


def stprint(
    data: Any,
    key: Optional[str] = None,
    indent: int = 0,
    file: Optional[io.TextIOBase] = None,
    indent_format: str = "  ",
    end_format: str = "\n",
    float_format: str = _DEFAULT_FLOAT_FORMAT,
    need_lock: bool = True,
    max_depth: int = 100,
) -> None:
    """
    Structure print.

    Example:
        >>> data = dict(a=np.zeros(shape=(10, 10)), b=3)
        >>> stprint(data)
        dict{
            a: ndarray(10, 10), dtype=float64
            b: 3
        }
    """
    if file is None:
        file = sys.stdout

    with stprint.locks.synchronized(file, need_lock):
        _process_data_recursive(data, indent, key, max_depth, indent_format, end_format, float_format, file)


stprint.locks = LockRegistry()


def stformat(data: Any, key: Optional[str] = None, indent: int = 0, max_depth: int = 100, **kwargs) -> str:
    return print2format(stprint)(data, key=key, indent=indent, need_lock=False, max_depth=max_depth, **kwargs)


def kvprint(
    data: dict,
    indent: int = 0,
    sep: str = " : ",
    end: str = "\n",
    max_key_len: Optional[int] = None,
    file: Optional[io.TextIOBase] = None,
    float_format: str = _DEFAULT_FLOAT_FORMAT,
    need_lock: bool = True,
) -> None:
    if len(data) == 0:
        return

    with kvprint.locks.synchronized(file, need_lock):
        keys = sorted(data.keys())
        lens = list(map(len, keys))
        # Replace if-else with ternary operator
        max_len = max_key_len if max_key_len is not None else max(lens)

        for k in keys:
            print("  " * indent, end="")
            print(
                k + " " * max(max_len - len(k), 0),
                format_printable_data(data[k], float_format=float_format),
                sep=sep,
                end=end,
                file=file,
                flush=True,
            )


kvprint.locks = LockRegistry()


def kvformat(data: dict, indent: int = 0, sep: str = " : ", end: str = "\n", max_key_len: Optional[int] = None) -> str:
    return print2format(kvprint)(data, indent=indent, sep=sep, end=end, max_key_len=max_key_len, need_lock=False)


class PrintToStringContext:
    __global_locks = LockRegistry()
    VALID_TARGETS: ClassVar[set[str]] = {"STDOUT", "STDERR"}

    def __init__(self, target: str = "STDOUT", stream: Optional[io.StringIO] = None, need_lock: bool = True):
        # Replace assert with ValueError
        if target not in self.VALID_TARGETS:
            raise ValueError(f"Target must be one of {self.VALID_TARGETS}")

        self._target = target
        self._need_lock = need_lock
        self._stream = stream if stream is not None else io.StringIO()
        self._stream_lock = threading.Lock()
        self._backup = None
        self._value = None

    def _swap(self, rhs: io.StringIO) -> io.StringIO:
        if self._target == "STDOUT":
            sys.stdout, rhs = rhs, sys.stdout
        else:
            sys.stderr, rhs = rhs, sys.stderr
        return rhs

    def __enter__(self) -> PrintToStringContext:
        if self._need_lock:
            self.__global_locks[self._target].acquire()
        self._backup = self._swap(self._stream)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._stream = self._swap(self._backup)
        if self._need_lock:
            self.__global_locks[self._target].release()

    def _ensure_value(self) -> None:
        if self._value is None:
            self._value = self._stream.getvalue()
            self._stream.close()

    def get(self) -> str:
        self._ensure_value()
        return self._value


def print_to_string(target: str = "STDOUT") -> PrintToStringContext:
    return PrintToStringContext(target, need_lock=True)


@contextlib.contextmanager
def print_to(print_func, target: str = "STDOUT", rstrip: bool = True):
    with PrintToStringContext(target, need_lock=True) as ctx:
        yield
    out_str = ctx.get()
    if rstrip:
        out_str = out_str.rstrip()
    print_func(out_str)


def print2format(print_func):
    def format_func(*args, **kwargs):
        f = io.StringIO()
        print_func(*args, file=f, **kwargs)
        value = f.getvalue()
        f.close()
        return value

    return format_func


def bubbletext(text: str, font: str = "cybermedium") -> str:
    """
    Uses pyfiglet to create bubble text.
    Args:
        font (str): default=cybermedium, other fonts include: cybersmall and
            cyberlarge.
    References:
        http://www.figlet.org/
    """
    try:
        import pyfiglet
    except ImportError:
        from jammy.logging import get_logger

        logger = get_logger()
        logger.debug("Missing pyfiglet when use bubbletext")
        return text
    else:
        bubble_text = pyfiglet.figlet_format(text, font=font)
        return bubble_text


class SupressPrintCTX:
    def __init__(self, stdout: bool = False, stderr: bool = False):
        self._stdout = stdout
        self._stderr = stderr
        self._stdout_backup = None
        self._stderr_backup = None

    def __enter__(self) -> None:
        if self._stdout:
            self._stdout_backup = sys.stdout
            sys.stdout = open(os.devnull, "w")
        if self._stderr:
            self._stderr_backup = sys.stderr
            sys.stderr = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self._stdout:
            sys.stdout.close()
            sys.stdout = self._stdout_backup
        if self._stderr:
            sys.stderr.close()
            sys.stderr = self._stderr_backup



import torch
import torch.nn as nn
import torch.nn.functional as F

class ToyModel(nn.Module):
    def __init__(self):
        super(ToyModel, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)
    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


import re
from collections import OrderedDict
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Tuple, TypeVar

import torch
from torch import nn
from torch.utils.hooks import RemovableHandle

# Type variables for generic types
T = TypeVar("T")
ModuleType = TypeVar("ModuleType", bound=nn.Module)  # pylint: disable=invalid-name

# Type aliases for better readability
FilterFn = Callable[[nn.Module], bool]
HookFn = Callable[[nn.Module, Tuple[torch.Tensor, ...], torch.Tensor], None]
RegisterHookFn = Callable[[str, nn.Module], RemovableHandle]
ForwardFn = Callable[[], T]


def all_layers_filter(_: nn.Module) -> bool:
    """Filter function to capture activations from all layers."""
    return True


@dataclass
class HookCaptureConfig:
    """Configuration for model hook capture.

    Attributes:
        remove_compiled_prefix: Whether to remove '_orig_mod.' prefix from compiled models.
        raise_on_hook_error: Whether to raise exceptions on hook removal errors.
        module_name_pattern: Optional regex pattern to filter module names.
    """

    remove_compiled_prefix: bool = True
    raise_on_hook_error: bool = False
    module_name_pattern: Optional[str] = None


def model_hook_capture(  # pylint: disable=too-many-branches
    model: ModuleType,
    forward_fn: ForwardFn[T],
    register_hook_fn: RegisterHookFn,
    filter_fn: FilterFn = all_layers_filter,
    config: Optional[HookCaptureConfig] = None,
) -> T:
    """Captures and processes intermediate activations in a PyTorch model using hooks.

    Args:
        model: PyTorch model to hook into.
        forward_fn: Function that runs the model forward pass.
        filter_fn: Function to filter which modules to hook.
        register_hook_fn: Function to register hooks for modules.
        config: Optional configuration for hook capture behavior.

    Returns:
        Results from forward_fn execution.

    Raises:
        RuntimeError: If hook removal fails and raise_on_hook_error is True.
        ValueError: If model has no modules matching the filter criteria.

    Example:
        ```python
        def custom_filter(module):
            return isinstance(module, nn.Conv2d)

        def custom_hook_register(name, module):
            def hook_fn(mod, inputs, output):
                print(f"{name}: {output.shape}")
            return module.register_forward_hook(hook_fn)

        results = model_hook_capture(
            model=my_model,
            forward_fn=lambda: model(input_tensor),
            filter_fn=custom_filter,
            register_hook_fn=custom_hook_register,
        )
        ```
    """
    if config is None:
        config = HookCaptureConfig()

    if not isinstance(model, nn.Module):
        raise TypeError(f"Expected nn.Module, got {type(model).__name__}")

    # Get module name mapping, handling compiled models
    module_to_name: Dict[nn.Module, str] = {}
    for name, module in model.named_modules():
        if config.remove_compiled_prefix:
            name = name.split("_orig_mod.")[-1]
        module_to_name[module] = name

    # Filter modules and store in OrderedDict to maintain order
    captured_modules = OrderedDict()
    for module, name in module_to_name.items():
        if not filter_fn(module):
            continue
        if config.module_name_pattern and not re.match(
            config.module_name_pattern, name
        ):
            continue
        captured_modules[name] = module

    if not captured_modules:
        raise ValueError(
            "No modules matched the filter criteria. Check your filter_fn and "
            "module_name_pattern if provided."
        )

    # Store hooks for cleanup
    hooks: List[RemovableHandle] = []

    try:
        # Register hooks for all captured modules
        for name, module in captured_modules.items():
            try:
                hook = register_hook_fn(name, module)
                if not isinstance(hook, RemovableHandle):
                    print(
                        f"Hook returned by register_hook_fn for module {name} is not a "
                        "RemovableHandle. Hook cleanup may fail."
                    )
                hooks.append(hook)
            except Exception as e:
                # log.error("Failed to register hook for module %s: %s", name, e)
                print(f"Failed to register hook for module {name}: {e}")
                # Clean up already registered hooks before re-raising
                _cleanup_hooks(hooks, raise_on_error=False)
                raise RuntimeError(f"Hook registration failed for module {name}") from e

        # Run the model
        try:
            results = forward_fn()
        except Exception as e:
            # log.error("Forward pass failed: %s", e)
            print(f"Forward pass failed: {e}")
            _cleanup_hooks(hooks, raise_on_error=False)
            raise

    finally:
        # Always clean up hooks
        _cleanup_hooks(hooks, raise_on_error=config.raise_on_hook_error)

    return results


def _cleanup_hooks(
    hooks: List[RemovableHandle],
    raise_on_error: bool = False,
) -> None:
    """Cleans up PyTorch hooks safely.

    Args:
        hooks: List of PyTorch removable handles to clean up.
        raise_on_error: Whether to raise exceptions on hook removal errors.

    Raises:
        RuntimeError: If hook removal fails and raise_on_error is True.
    """
    for hook in hooks:
        try:
            hook.remove()
        except Exception as e:  # pylint: disable=broad-except
            msg = f"Failed to remove hook: {e}"
            if raise_on_error:
                raise RuntimeError(msg) from e
            # log.warning(msg)
            print(f"warning: {msg}")
    hooks.clear()



# Copyright (c) OpenMMLab. All rights reserved.
import inspect
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator, Iterator, Optional, Tuple, Union

from jammy.io.path import is_filepath

from .backends import (
    BaseStorageBackend,
    Boto3Backend,
    HTTPBackend,
    LmdbBackend,
    LocalBackend,
    MemcachedBackend,
)


class HardDiskBackend(LocalBackend):
    """Raw hard disks storage backend."""

    @property
    def name(self):
        return self.__class__.__name__


class FileClient:
    """A general file client to access files in different backends.

    The client loads a file or text in a specified backend from its path
    and returns it as a binary or text file. There are two ways to choose a
    backend, the name of backend and the prefix of path. Although both of them
    can be used to choose a storage backend, ``backend`` has a higher priority
    that is if they are all set, the storage backend will be chosen by the
    backend argument. If they are all `None`, the disk backend will be chosen.
    Note that It can also register other backend accessor with a given name,
    prefixes, and backend class. In addition, We use the singleton pattern to
    avoid repeated object creation. If the arguments are the same, the same
    object will be returned.

    Warning:
        `FileClient` will be deprecated in future. Please use io functions
        in https://mmengine.readthedocs.io/en/latest/api/fileio.html#file-io

    Args:
        backend (str, optional): The storage backend type. Options are "disk",
            "memcached", "lmdb", "http" and "s3". Defaults to None.
        prefix (str, optional): The prefix of the registered storage backend.
            Options are "s3", "http", "https". Defaults to None.

    Examples:
        >>> # only set backend
        >>> file_client = FileClient(backend='s3')
        >>> # only set prefix
        >>> file_client = FileClient(prefix='s3')
        >>> # set both backend and prefix but use backend to choose client
        >>> file_client = FileClient(backend='s3', prefix='s3')
        >>> # if the arguments are the same, the same object is returned
        >>> file_client1 = FileClient(backend='s3')
        >>> file_client1 is file_client
        True

    Attributes:
        client (:obj:`BaseStorageBackend`): The backend object.
    """

    _backends = {
        "disk": HardDiskBackend,
        "memcached": MemcachedBackend,
        "lmdb": LmdbBackend,
        "s3": Boto3Backend,
        "http": HTTPBackend,
    }

    _prefix_to_backends: dict = {
        "s3": Boto3Backend,
        "s3": Boto3Backend,
        "http": HTTPBackend,
        "https": HTTPBackend,
    }

    _instances: dict = {}

    client: Any

    def __new__(cls, backend=None, prefix=None, **kwargs):
        if backend is None and prefix is None:
            backend = "disk"
        if backend is not None and backend not in cls._backends:
            raise ValueError(
                f"Backend {backend} is not supported. Currently supported ones"
                f" are {list(cls._backends.keys())}"
            )
        if prefix is not None and prefix not in cls._prefix_to_backends:
            raise ValueError(
                f"prefix {prefix} is not supported. Currently supported ones "
                f"are {list(cls._prefix_to_backends.keys())}"
            )

        # concatenate the arguments to a unique key for determining whether
        # objects with the same arguments were created
        arg_key = f"{backend}:{prefix}"
        for key, value in kwargs.items():
            arg_key += f":{key}:{value}"

        # if a backend was overridden, it will create a new object
        if arg_key in cls._instances:
            _instance = cls._instances[arg_key]
        else:
            # create a new object and put it to _instance
            _instance = super().__new__(cls)
            if backend is not None:
                _instance.client = cls._backends[backend](**kwargs)
            else:
                _instance.client = cls._prefix_to_backends[prefix](**kwargs)

            cls._instances[arg_key] = _instance

        return _instance

    @property
    def name(self):
        return self.client.name

    @property
    def allow_symlink(self):
        return self.client.allow_symlink

    @staticmethod
    def parse_uri_prefix(uri: Union[str, Path]) -> Optional[str]:
        """Parse the prefix of a uri.

        Args:
            uri (str | Path): Uri to be parsed that contains the file prefix.

        Examples:
            >>> FileClient.parse_uri_prefix('s3://path/of/your/file')
            's3'

        Returns:
            str | None: Return the prefix of uri if the uri contains '://' else
            ``None``.
        """
        assert is_filepath(uri)
        uri = str(uri)
        if "://" not in uri:
            return None
        else:
            prefix, _ = uri.split("://")
            # In the case of Boto3Backend, the prefix may contains the cluster
            # name like clusterName:s3
            if ":" in prefix:
                _, prefix = prefix.split(":")
            return prefix

    @classmethod
    def infer_client(
        cls,
        file_client_args: Optional[dict] = None,
        uri: Optional[Union[str, Path]] = None,
    ) -> "FileClient":
        """Infer a suitable file client based on the URI and arguments.

        Args:
            file_client_args (dict, optional): Arguments to instantiate a
                FileClient. Defaults to None.
            uri (str | Path, optional): Uri to be parsed that contains the file
                prefix. Defaults to None.

        Examples:
            >>> uri = 's3://path/of/your/file'
            >>> file_client = FileClient.infer_client(uri=uri)
            >>> file_client_args = {'backend': 's3'}
            >>> file_client = FileClient.infer_client(file_client_args)

        Returns:
            FileClient: Instantiated FileClient object.
        """
        assert file_client_args is not None or uri is not None
        if file_client_args is None:
            file_prefix = cls.parse_uri_prefix(uri)  # type: ignore
            return cls(prefix=file_prefix)
        else:
            return cls(**file_client_args)

    @classmethod
    def _register_backend(cls, name, backend, force=False, prefixes=None):
        if not isinstance(name, str):
            raise TypeError(
                "the backend name should be a string, " f"but got {type(name)}"
            )
        if not inspect.isclass(backend):
            raise TypeError(f"backend should be a class but got {type(backend)}")
        if not issubclass(backend, BaseStorageBackend):
            raise TypeError(
                f"backend {backend} is not a subclass of BaseStorageBackend"
            )
        if not force and name in cls._backends:
            raise KeyError(
                f"{name} is already registered as a storage backend, "
                'add "force=True" if you want to override it'
            )

        if name in cls._backends and force:
            for arg_key, instance in list(cls._instances.items()):
                if isinstance(instance.client, cls._backends[name]):
                    cls._instances.pop(arg_key)
        cls._backends[name] = backend

        if prefixes is not None:
            if isinstance(prefixes, str):
                prefixes = [prefixes]
            else:
                assert isinstance(prefixes, (list, tuple))
            for prefix in prefixes:
                if prefix not in cls._prefix_to_backends:
                    cls._prefix_to_backends[prefix] = backend
                elif (prefix in cls._prefix_to_backends) and force:
                    overridden_backend = cls._prefix_to_backends[prefix]
                    for arg_key, instance in list(cls._instances.items()):
                        if isinstance(instance.client, overridden_backend):
                            cls._instances.pop(arg_key)
                else:
                    raise KeyError(
                        f"{prefix} is already registered as a storage backend,"
                        ' add "force=True" if you want to override it'
                    )

    @classmethod
    def register_backend(cls, name, backend=None, force=False, prefixes=None):
        """Register a backend to FileClient.

        This method can be used as a normal class method or a decorator.

        .. code-block:: python

            class NewBackend(BaseStorageBackend):

                def get(self, filepath):
                    return filepath

                def get_text(self, filepath):
                    return filepath

            FileClient.register_backend('new', NewBackend)

        or

        .. code-block:: python

            @FileClient.register_backend('new')
            class NewBackend(BaseStorageBackend):

                def get(self, filepath):
                    return filepath

                def get_text(self, filepath):
                    return filepath

        Args:
            name (str): The name of the registered backend.
            backend (class, optional): The backend class to be registered,
                which must be a subclass of :class:`BaseStorageBackend`.
                When this method is used as a decorator, backend is None.
                Defaults to None.
            force (bool, optional): Whether to override the backend if the name
                has already been registered. Defaults to False.
            prefixes (str or list[str] or tuple[str], optional): The prefixes
                of the registered storage backend. Defaults to None.
                `New in version 1.3.15.`
        """
        if backend is not None:
            cls._register_backend(name, backend, force=force, prefixes=prefixes)
            return

        def _register(backend_cls):
            cls._register_backend(name, backend_cls, force=force, prefixes=prefixes)
            return backend_cls

        return _register

    def get(self, filepath: Union[str, Path]) -> Union[bytes, memoryview]:
        """Read data from a given ``filepath`` with 'rb' mode.

        Note:
            There are two types of return values for ``get``, one is ``bytes``
            and the other is ``memoryview``. The advantage of using memoryview
            is that you can avoid copying, and if you want to convert it to
            ``bytes``, you can use ``.tobytes()``.

        Args:
            filepath (str or Path): Path to read data.

        Returns:
            bytes | memoryview: Expected bytes object or a memory view of the
            bytes object.
        """
        return self.client.get(filepath)

    def get_text(self, filepath: Union[str, Path], encoding="utf-8") -> str:
        """Read data from a given ``filepath`` with 'r' mode.

        Args:
            filepath (str or Path): Path to read data.
            encoding (str): The encoding format used to open the ``filepath``.
                Defaults to 'utf-8'.

        Returns:
            str: Expected text reading from ``filepath``.
        """
        return self.client.get_text(filepath, encoding)

    def put(self, obj: bytes, filepath: Union[str, Path]) -> None:
        """Write data to a given ``filepath`` with 'wb' mode.

        Note:
            ``put`` should create a directory if the directory of ``filepath``
            does not exist.

        Args:
            obj (bytes): Data to be written.
            filepath (str or Path): Path to write data.
        """
        self.client.put(obj, filepath)

    def put_text(self, obj: str, filepath: Union[str, Path]) -> None:
        """Write data to a given ``filepath`` with 'w' mode.

        Note:
            ``put_text`` should create a directory if the directory of
            ``filepath`` does not exist.

        Args:
            obj (str): Data to be written.
            filepath (str or Path): Path to write data.
            encoding (str, optional): The encoding format used to open the
                `filepath`. Defaults to 'utf-8'.
        """
        self.client.put_text(obj, filepath)

    def remove(self, filepath: Union[str, Path]) -> None:
        """Remove a file.

        Args:
            filepath (str, Path): Path to be removed.
        """
        self.client.remove(filepath)

    def exists(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path exists.

        Args:
            filepath (str or Path): Path to be checked whether exists.

        Returns:
            bool: Return ``True`` if ``filepath`` exists, ``False`` otherwise.
        """
        return self.client.exists(filepath)

    def isdir(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a directory.

        Args:
            filepath (str or Path): Path to be checked whether it is a
                directory.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a directory,
            ``False`` otherwise.
        """
        return self.client.isdir(filepath)

    def isfile(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a file.

        Args:
            filepath (str or Path): Path to be checked whether it is a file.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a file, ``False``
            otherwise.
        """
        return self.client.isfile(filepath)

    def join_path(
        self, filepath: Union[str, Path], *filepaths: Union[str, Path]
    ) -> str:
        r"""Concatenate all file paths.

        Join one or more filepath components intelligently. The return value
        is the concatenation of filepath and any members of \*filepaths.

        Args:
            filepath (str or Path): Path to be concatenated.

        Returns:
            str: The result of concatenation.
        """
        return self.client.join_path(filepath, *filepaths)

    @contextmanager
    def get_local_path(
        self, filepath: Union[str, Path]
    ) -> Generator[Union[str, Path], None, None]:
        """Download data from ``filepath`` and write the data to local path.

        ``get_local_path`` is decorated by :meth:`contxtlib.contextmanager`. It
        can be called with ``with`` statement, and when exists from the
        ``with`` statement, the temporary path will be released.

        Note:
            If the ``filepath`` is a local path, just return itself.

        .. warning::
            ``get_local_path`` is an experimental interface that may change in
            the future.

        Args:
            filepath (str or Path): Path to be read data.

        Examples:
            >>> file_client = FileClient(prefix='s3')
            >>> with file_client.get_local_path('s3://bucket/abc.jpg') as path:
            ...     # do something here

        Yields:
            Iterable[str]: Only yield one path.
        """
        with self.client.get_local_path(str(filepath)) as local_path:
            yield local_path

    def list_dir_or_file(  # pylint: disable=too-many-arguments
        self,
        dir_path: Union[str, Path],
        list_dir: bool = True,
        list_file: bool = True,
        suffix: Optional[Union[str, Tuple[str]]] = None,
        recursive: bool = False,
    ) -> Iterator[str]:
        """Scan a directory to find the interested directories or files in
        arbitrary order.

        Note:
            :meth:`list_dir_or_file` returns the path relative to ``dir_path``.

        Args:
            dir_path (str | Path): Path of the directory.
            list_dir (bool): List the directories. Defaults to True.
            list_file (bool): List the path of files. Defaults to True.
            suffix (str or tuple[str], optional):  File suffix
                that we are interested in. Defaults to None.
            recursive (bool): If set to True, recursively scan the
                directory. Defaults to False.

        Yields:
            Iterable[str]: A relative path to ``dir_path``.
        """
        yield from self.client.list_dir_or_file(
            dir_path, list_dir, list_file, suffix, recursive
        )



from collections import abc
from typing import Any, Type, Union


def has_method(obj, method):
    return hasattr(obj, method) and callable(getattr(obj, method))


def iter_cast(inputs, dst_type, return_type=None):
    """Cast elements of an iterable object into some type.

    Args:
        inputs (Iterable): The input object.
        dst_type (type): Destination type.
        return_type (type, optional): If specified, the output object will be
            converted to this type, otherwise an iterator.

    Returns:
        iterator or specified type: The converted object.
    """
    if not isinstance(inputs, abc.Iterable):
        raise TypeError("inputs must be an iterable object")
    if not isinstance(dst_type, type):
        raise TypeError('"dst_type" must be a valid type')

    out_iterable = map(dst_type, inputs)

    if return_type is None:
        return out_iterable
    else:
        return return_type(out_iterable)


def list_cast(inputs, dst_type):
    """Cast elements of an iterable object into a list of some type.

    A partial method of :func:`iter_cast`.
    """
    return iter_cast(inputs, dst_type, return_type=list)


def tuple_cast(inputs, dst_type):
    """Cast elements of an iterable object into a tuple of some type.

    A partial method of :func:`iter_cast`.
    """
    return iter_cast(inputs, dst_type, return_type=tuple)


def is_seq_of(
    seq: Any, expected_type: Union[Type, tuple], seq_type: Type = None
) -> bool:
    """Check whether it is a sequence of some type.

    Args:
        seq (Sequence): The sequence to be checked.
        expected_type (type or tuple): Expected type of sequence items.
        seq_type (type, optional): Expected sequence type. Defaults to None.

    Returns:
        bool: Return True if ``seq`` is valid else False.

    Examples:
        >>> from mmengine.utils import is_seq_of
        >>> seq = ['a', 'b', 'c']
        >>> is_seq_of(seq, str)
        True
        >>> is_seq_of(seq, int)
        False
    """
    if seq_type is None:
        exp_seq_type = abc.Sequence
    else:
        assert isinstance(seq_type, type)
        exp_seq_type = seq_type
    if not isinstance(seq, exp_seq_type):
        return False
    for item in seq:
        if not isinstance(item, expected_type):
            return False
    return True


def is_list_of(seq, expected_type):
    """Check whether it is a list of some type.

    A partial method of :func:`is_seq_of`.
    """
    return is_seq_of(seq, expected_type, seq_type=list)


def is_tuple_of(seq, expected_type):
    """Check whether it is a tuple of some type.

    A partial method of :func:`is_seq_of`.
    """
    return is_seq_of(seq, expected_type, seq_type=tuple)



import itertools
from collections.abc import Iterable
from functools import partial


def _args_single_param(item, _format="args"):
    key, value = item
    rtn = []
    if not isinstance(value, Iterable):
        value = [value]
    fist_element = value[0] if isinstance(value, list) else value
    if _format == "args":
        if isinstance(fist_element, bool):
            for cur_value in value:
                arg = "--" + key if cur_value else "--no_" + key
                rtn.append(arg)
            return rtn
        for i in value:
            rtn.append(f"--{key} {i}")
    elif _format == "hydra":
        for i in value:
            rtn.append(f"{key}={i}")
    else:
        raise RuntimeError(f"{_format} not supported yet")

    return rtn


def param_sweep(prefix=None, **kwargs):
    """generate cmds

    :param prefix: defaults to None
    : conditional "_format" -> args, hydra
    """
    _format = "args"
    if "_format" in kwargs:
        _format = kwargs.get("_format")
        del kwargs["_format"]
    assert _format in ["args", "hydra"]
    args_list = list(map(partial(_args_single_param, _format=_format), kwargs.items()))
    if prefix is not None:
        args_list.insert(0, [prefix])
    permutation = list(itertools.product(*args_list))
    return [" ".join(item) for item in permutation]



# Copyright (c) OpenMMLab. All rights reserved.
# pylint: disable=import-error, import-outside-toplevel
from pathlib import Path
from typing import Union

from .base import BaseStorageBackend


class LmdbBackend(BaseStorageBackend):
    """Lmdb storage backend.

    Args:
        db_path (str): Lmdb database path.
        readonly (bool): Lmdb environment parameter. If True, disallow any
            write operations. Defaults to True.
        lock (bool): Lmdb environment parameter. If False, when concurrent
            access occurs, do not lock the database. Defaults to False.
        readahead (bool): Lmdb environment parameter. If False, disable the OS
            filesystem readahead mechanism, which may improve random read
            performance when a database is larger than RAM. Defaults to False.
        **kwargs: Keyword arguments passed to `lmdb.open`.

    Attributes:
        db_path (str): Lmdb database path.
    """

    def __init__(self, db_path, readonly=True, lock=False, readahead=False, **kwargs):
        try:
            import lmdb  # noqa: F401 # pylint: disable=unused-import
        except ImportError:
            raise ImportError('Please run "pip install lmdb" to enable LmdbBackend.')

        self.db_path = str(db_path)
        self.readonly = readonly
        self.lock = lock
        self.readahead = readahead
        self.kwargs = kwargs
        self._client = None

    def get(self, filepath: Union[str, Path]) -> bytes:
        """Get values according to the filepath.

        Args:
            filepath (str or Path): Here, filepath is the lmdb key.

        Returns:
            bytes: Expected bytes object.

        Examples:
            >>> backend = LmdbBackend('path/to/lmdb')
            >>> backend.get('key')
            b'hello world'
        """
        if self._client is None:
            self._client = self._get_client()

        filepath = str(filepath)
        with self._client.begin(write=False) as txn:
            value_buf = txn.get(filepath.encode("ascii"))
        return value_buf

    def get_text(self, filepath, encoding=None):
        raise NotImplementedError

    def _get_client(self):
        import lmdb

        return lmdb.open(
            self.db_path,
            readonly=self.readonly,
            lock=self.lock,
            readahead=self.readahead,
            **self.kwargs
        )

    def __del__(self):
        if self._client is not None:
            self._client.close()



# pylint: skip-file
from jammy.storage.kv.lmdb import LMDBKVStore


def main():
    kv = LMDBKVStore("/tmp/test_1.lmdb", readonly=False)

    kv.put("a", 1)
    kv.put("b", 2)

    assert "a" in kv and kv["a"] == 1
    assert "b" in kv and kv["b"] == 2
    assert "c" not in kv

    for k in kv.keys():
        print(k, kv[k])


if __name__ == "__main__":
    main()



from .deep_to import *
from .device import *
from .to import *


def get_torch_info():
    torch_flag_list = [
        "torch.backends.cuda.is_built",
        "torch.backends.cuda.matmul.allow_tf32",
        "torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction",
        "torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction",
        "torch.backends.cudnn.allow_tf32",
        "torch.backends.cudnn.benchmark",
        "torch.backends.cudnn.enabled",
        "torch.backends.cudnn.deterministic",
        "torch.backends.cudnn.version",
        "torch.get_default_dtype",
    ]
    torch_info = {}
    for flag in torch_flag_list:
        value = eval(flag)  # pylint: disable=eval-used
        if callable(value):
            torch_info[flag] = value()
        else:
            torch_info[flag] = value
    return torch_info



import copy

import torch
from torch import nn

from jamtorch.logging import get_logger

__all__ = ["EMA"]


class EMA:
    def __init__(
        self, beta, num_warm, num_every, model=None, forget_resume=False
    ):  # pylint: disable=too-many-arguments
        self.beta = beta
        self.num_warm = num_warm
        self.num_every = num_every
        self.cnt = 0
        if isinstance(model, nn.DataParallel):
            model = model.module
        self.ema_delay_model = copy.deepcopy(model)
        self.forget_resume = forget_resume

    def forward(self, *args, **kwargs):
        return self.ema_delay_model(*args, **kwargs)

    def update_parameters(self, new_model):
        if hasattr(new_model, "module"):
            new_model = new_model.module
        self.cnt += 1
        if self.cnt < self.num_warm:
            return
        if self.model is None:
            self.model = new_model
        if self.cnt == self.num_warm:
            self.ema_delay_model.load_state_dict(new_model.state_dict())
            return
        with torch.no_grad():
            if self.cnt % self.num_every == 0:
                for old_param, new_param in zip(
                    self.model.parameters(), new_model.parameters()
                ):
                    old_weight, new_weight = old_param.data, new_param.data
                    old_param.data = (
                        old_weight * self.beta + (1 - self.beta) * new_weight
                    )

    @property
    def model(self):
        return self.ema_delay_model

    @model.setter
    def model(self, model):
        if isinstance(model, nn.DataParallel):
            model = model.module
        self.ema_delay_model = copy.deepcopy(model)

    def state_dict(self):
        if self.ema_delay_model is None:
            model_params = None
        else:
            model_params = self.ema_delay_model.state_dict()
        return {
            "cnt": self.cnt,
            "model": model_params,
            "num_warm": self.num_warm,
            "num_every": self.num_every,
            "beta": self.beta,
        }

    def copy_to(self, model):  # pylint: disable= unused-argument
        # return self.ema_delay_model
        model.load_state_dict(self.ema_delay_model.state_dict())

    def load_state_dict(self, ema_dict):
        logger = get_logger()
        self.cnt = ema_dict.get("cnt") or 0
        self.num_warm = ema_dict.get("num_warm") or 0
        if not self.forget_resume:
            if self.ema_delay_model is not None and ema_dict.get("model") is not None:
                self.ema_delay_model.load_state_dict(ema_dict.get("model"))
        self.beta = ema_dict.get("beta") or 0.9
        self.num_every = ema_dict.get("num_every") or 1
        logger.critical("=====> Loading EMA finish =====>")



import os
import pathlib
import resource
import sys

from jammy.utils.env import jam_getenv, jam_is_debug

from .meta import run_once


def release_syslim():
    if jam_getenv("SYSLIM", default="n", type="bool"):
        sys.setrecursionlimit(1000000)
        try:
            slim = 65536 * 1024
            resource.setrlimit(resource.RLIMIT_STACK, (slim, slim))
        except ValueError:
            pass


def tune_opencv():
    os.environ["OPENCV_OPENCL_RUNTIME"] = ""


def enable_ipdb():
    if jam_is_debug():
        if jam_getenv("IMPORT_ALL", "true", "bool"):
            from jammy.utils.debug import hook_exception_ipdb

            hook_exception_ipdb()


def init_main():
    release_syslim()
    tune_opencv()
    enable_ipdb()
    main_path()


@run_once
def main_path():
    if jam_getenv("proj_path") is None:
        os.environ["JAM_PROJ_PATH"] = str(pathlib.Path().absolute())
        return pathlib.Path().absolute()
    return jam_getenv("proj_path")



import pytest
import torch

from jamtorch.nn.gpt import GPT, GPTConfig

from .hooks_apps import PerLayerSpec, layer_sepc_to_df, terok_processor
from .hooks_core import all_layers_filter, model_hook_capture


def test_forward():
    gpt_config = GPTConfig()
    gpt_model = GPT(gpt_config).cuda()
    per_layer_spec = PerLayerSpec(process_fn=terok_processor)

    def _fn():
        return gpt_model(torch.randint(0, gpt_config.vocab_size, (2, 128)).cuda())

    _ = model_hook_capture(
        gpt_model,
        _fn,
        per_layer_spec.register_hook,
        all_layers_filter,
        None,
    )

    layer_spec = per_layer_spec.results
    print(layer_sepc_to_df(layer_spec))


if __name__ == "__main__":
    pytest.main([__file__])



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : _lazy_bind.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import functools

from .rng import get_default_rng, _rng

__all__ = [
    # Utility functions
    'random_sample',        # Uniformly distributed floats over ``[0, 1)``.
    'bytes',                # Uniformly distributed random bytes.
    'random_integers',      # Uniformly distributed integers in a given range.
    'permutation',          # Randomly permute a sequence / generate a random sequence.
    'shuffle',              # Randomly permute a sequence in place.
    'seed',                 # Seed the random number generator.
    'choice',               # Random sample from 1-D array.

    # Compatibility functions
    'rand',                 # Uniformly distributed values.
    'randn',                # Normally distributed values.
    'randint',              # Uniformly distributed integers in a given range.

    # Univariate distributions
    'beta',                 # Beta distribution over ``[0, 1]``.
    'binomial',             # Binomial distribution.
    'chisquare',            # :math:`\\chi^2` distribution.
    'exponential',          # Exponential distribution.
    'f',                    # F (Fisher-Snedecor) distribution.
    'gamma',                # Gamma distribution.
    'geometric',            # Geometric distribution.
    'gumbel',               # Gumbel distribution.
    'hypergeometric',       # Hypergeometric distribution.
    'laplace',              # Laplace distribution.
    'logistic',             # Logistic distribution.
    'lognormal',            # Log-normal distribution.
    'logseries',            # Logarithmic series distribution.
    'negative_binomial',    # Negative binomial distribution.
    'noncentral_chisquare', # Non-central chi-square distribution.
    'noncentral_f',         # Non-central F distribution.
    'normal',               # Normal / Gaussian distribution.
    'pareto',               # Pareto distribution.
    'poisson',              # Poisson distribution.
    'power',                # Power distribution.
    'rayleigh',             # Rayleigh distribution.
    'triangular',           # Triangular distribution.
    'uniform',              # Uniform distribution.
    'vonmises',             # Von Mises circular distribution.
    'wald',                 # Wald (inverse Gaussian) distribution.
    'weibull',              # Weibull distribution.
    'zipf',                 # Zipf's distribution over ranked data.

    # Multivariate distributions
    'dirichlet',            # Multivariate generalization of Beta distribution.
    'multinomial',          # Multivariate generalization of the binomial distribution.
    'multivariate_normal',  # Multivariate generalization of the normal distribution.

    # Standard distributions
    'standard_cauchy',      # Standard Cauchy-Lorentz distribution.
    'standard_exponential', # Standard exponential distribution.
    'standard_gamma',       # Standard Gamma distribution.
    'standard_normal',      # Standard normal distribution.
    'standard_t',           # Standard Student's t-distribution.

    # Internal functions
    'get_state',            # Get tuple representing internal state of generator.
    'set_state',            # Set state of generator.

    # Customized utility functions
    'choice_list',          # Choose an element from a python list.
    'shuffle_list',         # Shuffle an python list.
    'shuffle_multi',        # Shuffle multiple arrays simultaneously.
]


def _lazy_bind(instance_func):
    func_name = instance_func.__name__

    @functools.wraps(instance_func)
    def wrapped(*args, **kwargs):
        func = getattr(get_default_rng(), func_name)
        return func(*args, **kwargs)

    return wrapped


# Utility functions
random_sample = _lazy_bind(_rng.random_sample)
bytes = _lazy_bind(_rng.bytes)
random_integers = _lazy_bind(_rng.random_integers)
permutation = _lazy_bind(_rng.permutation)
shuffle = _lazy_bind(_rng.shuffle)
seed = _lazy_bind(_rng.seed)
choice = _lazy_bind(_rng.choice)
# Compatibility functions
rand = _lazy_bind(_rng.rand)
randn = _lazy_bind(_rng.randn)
randint = _lazy_bind(_rng.randint)
# Univariate distributions
beta = _lazy_bind(_rng.beta)
binomial = _lazy_bind(_rng.binomial)
chisquare = _lazy_bind(_rng.chisquare)
exponential = _lazy_bind(_rng.exponential)
f = _lazy_bind(_rng.f)
gamma = _lazy_bind(_rng.gamma)
geometric = _lazy_bind(_rng.geometric)
gumbel = _lazy_bind(_rng.gumbel)
hypergeometric = _lazy_bind(_rng.hypergeometric)
laplace = _lazy_bind(_rng.laplace)
logistic = _lazy_bind(_rng.logistic)
lognormal = _lazy_bind(_rng.lognormal)
logseries = _lazy_bind(_rng.logseries)
negative_binomial = _lazy_bind(_rng.negative_binomial)
noncentral_chisquare = _lazy_bind(_rng.noncentral_chisquare)
noncentral_f = _lazy_bind(_rng.noncentral_f)
normal = _lazy_bind(_rng.normal)
pareto = _lazy_bind(_rng.pareto)
poisson = _lazy_bind(_rng.poisson)
power = _lazy_bind(_rng.power)
rayleigh = _lazy_bind(_rng.rayleigh)
triangular = _lazy_bind(_rng.triangular)
uniform = _lazy_bind(_rng.uniform)
vonmises = _lazy_bind(_rng.vonmises)
wald = _lazy_bind(_rng.wald)
weibull = _lazy_bind(_rng.weibull)
zipf = _lazy_bind(_rng.zipf)
# Multivariate distributions
dirichlet = _lazy_bind(_rng.dirichlet)
multinomial = _lazy_bind(_rng.multinomial)
multivariate_normal = _lazy_bind(_rng.multivariate_normal)
# Standard distributions
standard_cauchy = _lazy_bind(_rng.standard_cauchy)
standard_exponential = _lazy_bind(_rng.standard_exponential)
standard_gamma = _lazy_bind(_rng.standard_gamma)
standard_normal = _lazy_bind(_rng.standard_normal)
standard_t = _lazy_bind(_rng.standard_t)
# Internal functions
get_state = _lazy_bind(_rng.get_state)
set_state = _lazy_bind(_rng.set_state)
# Customized utility functions
choice_list = _lazy_bind(_rng.choice_list)
shuffle_list = _lazy_bind(_rng.shuffle_list)
shuffle_multi = _lazy_bind(_rng.shuffle_multi)

# Some aliases:
ranf = random = sample = random_sample
__all__.extend(['ranf', 'random', 'sample'])



import jammy
from jammy.utils.notifier import jam_notifier

parser = jammy.cli.JamArgumentParser()
parser.add_argument(
    "text", action="store", type=str, nargs="+", help="The text to send."
)

args = parser.parse_args()

if __name__ == "__main__":
    words = [cur_text.replace('"', '\\"') for cur_text in args.text]
    msg = " ".join(words)
    jam_notifier.notify(msg)



import base64
from io import BytesIO
from pathlib import Path
from typing import Optional, Union

from IPython.display import HTML, display


def display_video(  # pylint: disable=too-many-arguments
    file: Union[str, Path, bytes, bytearray, BytesIO],
    width: Optional[int] = 640,
    height: Optional[int] = 480,
    autoplay: bool = False,
    loop: bool = False,
    muted: bool = False,
) -> None:
    """
    Display a video file in a Jupyter Notebook using a data URL.

    Args:
        file (Union[str, Path, bytes, bytearray, BytesIO]): The video file to display.
            Can be a file path (str or Path) or video data (bytes, bytearray, or BytesIO).
        width (Optional[int]): The width of the video player in pixels. Set to None for default size.
        height (Optional[int]): The height of the video player in pixels. Set to None for default size.
        autoplay (bool): Whether the video should start playing automatically.
        loop (bool): Whether the video should loop when it reaches the end.
        muted (bool): Whether the video should be muted by default.

    Raises:
        FileNotFoundError: If the specified file path does not exist.
        ValueError: If the input type is not supported.
    """
    # Handle different input types
    if isinstance(file, (str, Path)):
        file_path = Path(file)
        if not file_path.exists():
            raise FileNotFoundError(f"File not found: {file_path}")
        with open(file_path, "rb") as f:
            video = f.read()
    elif isinstance(file, (bytes, bytearray)):
        video = bytes(file)
    elif isinstance(file, BytesIO):
        video = file.getvalue()
    else:
        raise ValueError(f"Unsupported input type: {type(file)}")

    # Encode video data as base64
    video_url = f"data:video/mp4;base64,{base64.b64encode(video).decode('utf-8')}"

    # Prepare video attributes
    attributes = []
    if width is not None:
        attributes.append(f'width="{width}"')
    if height is not None:
        attributes.append(f'height="{height}"')
    if autoplay:
        attributes.append("autoplay")
    if loop:
        attributes.append("loop")
    if muted:
        attributes.append("muted")
    attributes.append("controls")

    # Create HTML for video player
    video_html = f"""
    <video {' '.join(attributes)}>
        <source src="{video_url}" type="video/mp4">
        Your browser does not support the video tag.
    </video>
    """
    display(HTML(video_html))



import os
import os.path as osp

import torch
from omegaconf import DictConfig, OmegaConf
from torch.utils.data import DataLoader

from jammy.event import SimpleEventRegistry
from jammy.utils.enum import JamEnum
from jamtorch.io import attr_dict, load_ckpt, save_ckpt
from jamtorch.logging import get_logger

from . import trainer_fn

logger = get_logger()

__all__ = ["EvalState", "GeneticTrainer"]


class EvalState(JamEnum):
    NO = 1  # pylint: disable= invalid-name
    ITER = 2
    EPOCH = 3


class GeneticTrainer:  # pylint: disable=too-many-instance-attributes
    def __init__(self, cfg, loss_fn):
        self._device = None
        self.train_set = None
        self.train_loader = None
        self.val_set = None
        self.val_loader = None
        self.model = None
        self.optimizer = None
        self.lr_scheduler = None
        self.loss_fn = loss_fn

        self.epoch_cnt = 0
        self.iter_cnt = 0
        self.best_loss = float("inf")
        self.ckpt_dir = getattr(cfg, "ckpt_dir") or os.getcwd()
        self._states = ["epoch_cnt", "iter_cnt", "best_loss"]
        self.latest_ckpt = None

        self._cfg = cfg
        self.eval_epoch = cfg.get("eval_epoch") or 1
        self.eval_iter = cfg.get("eval_iter") or -1
        self.ratio_forback = cfg.get("ratio_forback") or 1
        self.use_amp = cfg.get("use_amp") or False
        self.amp_scaler = torch.cuda.amp.GradScaler(enabled=self.use_amp)
        if "gpu" in self._cfg:
            self.device = cfg["gpu"]

        self._event_manager = SimpleEventRegistry(
            {
                "epoch:start",
                "epoch:end",
                "epoch:before",
                "epoch:after",
                "step:start",
                "step:end",
                "step:summary",
                "forward:before",
                "forward:after",
                "backward:before",
                "backward:after",
                "val:start",
                "val:end",
                "val:step",  # quite strange naming and design
                "trainer:export",
                "trainer:load",
            }
        )
        if cfg.clip_grad and cfg.clip_grad > 0:
            trainer_fn.register_grad_clip(self, cfg.clip_grad)

    @property
    def mmodel(self):
        if hasattr(self.model, "module"):
            return self.model.module
        return self.model

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, gpu):
        if isinstance(gpu, int):
            self._device = torch.device(f"cuda:{gpu}")
        elif isinstance(gpu, str) and len(gpu) < 2:
            self._device = torch.device(f"cuda:{gpu}")
        else:
            self._device = gpu

    def set_dataset(self, train_set, val_set=None, loader_cfg=None):
        self.train_set, self.val_set = train_set, val_set
        if loader_cfg is not None:
            self.quick_loader(loader_cfg)

    def set_dataloader(self, train_loader, val_loader):
        self.train_loader = train_loader
        self.val_loader = val_loader

    def quick_loader(self, loader_cfg):
        if isinstance(loader_cfg, DictConfig):
            loader_cfg = OmegaConf.to_container(loader_cfg, resolve=True)
        if self.train_set is not None:
            self.train_loader = DataLoader(self.train_set, **loader_cfg)
        if self.val_set is not None:
            self.val_loader = DataLoader(self.val_set, **loader_cfg)

    def set_model_optim(self, model, optimizer=None):
        if next(model.parameters()).device == torch.device("cpu"):
            model = model.to(self.device)
        self.model, self.optimizer = model, optimizer
        # resume will loss one epoch, it is acceptable
        if self._cfg.resume:
            self.load_ckpt(self._cfg.ckpt)

    def export_env(self):
        states = {key: getattr(self, key) for key in self._states}
        return states

    def load_env(self, cfg):
        for key, value in cfg.items():
            if value is None:
                continue
            if key in self._states:
                setattr(self, key, value)

    def load_ckpt(self, filename="checkpoint"):
        state = load_ckpt(self.device, filename)
        self._impl_load_ckpt(state)
        self.load_env(state["env"])
        self.trigger_event("trainer:load", self, state)

    def _impl_load_ckpt(self, state):
        msg_model = self.model.load_state_dict(state["model"])
        logger.critical(f"load model ckpt {msg_model}")
        if "optimizer" in state:
            msg_optimizer = self.optimizer.load_state_dict(state["optimizer"])
            logger.critical(f"load optimizer ckpt {msg_optimizer}")
        else:
            logger.critical("no optimizer in state")
        self._impl_load_amp_scaler(state)

    def _impl_load_amp_scaler(self, state):
        if self.use_amp:
            if "amp_scaler" in state and len(state["amp_scaler"]) == 0:
                self.amp_scaler.load_state_dict(state["amp_scaler"])
            else:
                logger.critical("enabled amp but amp_scaler not found")

    def save_ckpt(self, val_loss=float("inf")):
        is_best = val_loss < self.best_loss
        self.best_loss = min(val_loss, self.best_loss)
        env_state = self.export_env()
        state_dict = {"env": env_state}
        ckpt_path = osp.join(self.ckpt_dir, f"ckpt-{self.iter_cnt}")
        self.latest_ckpt = ckpt_path
        best_path = osp.join(self.ckpt_dir, "best")
        state_dict.update(self._impl_save_ckpt())
        self.trigger_event("trainer:export", self, state_dict)

        save_ckpt(state_dict, is_best, ckpt_path, best_path)

    def _impl_save_ckpt(self):
        if self.use_amp:
            return attr_dict(self, ["model", "optimizer", "amp_scaler"])
        return attr_dict(self, ["model", "optimizer"])

    def register_event(self, name, callback, verbose=True):
        """
        "epoch:start/end/before/after",(trainer)
        "step:start/end",(trainer)
        "step:summary",(trainer, loss, monitors, cmdviz)

        "forward:before",(trainer, feed_dict)
        "forward:after",(trainer, feed_dict, loss, monitors, cmdviz)
        "backward:before",(trainer, feed_dict, loss, monitors, cmdviz)
        "backward:after",(trainer, feed_dict, loss, monitors, cmdviz)

        "val:start/end",(trainer)
        "val:step",(trainer, feed_dict, loss, monitors, cmdviz)

        "trainer:export",(trainer, state_dict)
        "trainer:load",(trainer, state)
        """

        if verbose:
            logger.info(
                "Register trainer event: name={}, callback={}.".format(
                    name, callback.__module__ + "." + callback.__name__
                )
            )
        else:
            logger.debug(
                "Register trainer event: name={}, callback={}.".format(
                    name, callback.__module__ + "." + callback.__name__
                )
            )
        self._event_manager.register(name, callback)

    def trigger_event(self, name, *args, **kwargs):
        self._event_manager.trigger(name, *args, **kwargs)

    def eval(self):
        self.model.eval()
        self.trigger_event("val:start", self)
        with torch.no_grad():
            for _, batch in enumerate(self.val_loader):
                with torch.cuda.amp.autocast(enabled=self.use_amp):
                    loss, monitor, cmdviz_dict = self.loss_fn(
                        self, batch, is_train=False
                    )
                self.trigger_event("val:step", self, batch, loss, monitor, cmdviz_dict)
        self.trigger_event("val:end", self)

    def eval_state(self):
        if self.val_loader is None:
            return EvalState.NO

        if self.eval_iter > 0 and (self.iter_cnt % self.eval_iter) == 0:
            if self.eval_epoch > 0 and ((self.epoch_cnt + 1) % self.eval_epoch) == 0:
                return EvalState.EPOCH
            return EvalState.ITER

        if self.iter_cnt % len(self.train_loader) == 0:
            if self.eval_epoch > 0 and ((self.epoch_cnt + 1) % self.eval_epoch) == 0:
                return EvalState.EPOCH
        return EvalState.NO

    def train(self):
        self.trigger_event("epoch:start", self)
        self.optimizer.zero_grad()
        for self.epoch_cnt in range(self.epoch_cnt, self._cfg.epochs):
            self.trigger_event("epoch:before", self)
            for _, batch in enumerate(self.train_loader):
                self.iter_cnt += 1
                self.trigger_event("step:start", self)
                self.train_step(batch)
                self.trigger_event("step:end", self)
                state = self.eval_state()
                if state is EvalState.NO:
                    self.monitor_update()
                elif state is EvalState.ITER:
                    self._eval_iter()
                    self.monitor_update()
            if state is EvalState.EPOCH:
                self._eval_epoch()
                self.monitor_update()
            self.trigger_event("epoch:after", self)
        self.trigger_event("epoch:end", self)
        self.eval()
        return self.best_loss

    def train_step(self, feed_dict):
        self.trigger_event("forward:before", self, feed_dict)
        with torch.cuda.amp.autocast(enabled=self.use_amp):
            loss, monitors, cmdviz_dict = self.loss_fn(self, feed_dict, is_train=True)

        self.trigger_event(
            "forward:after", self, feed_dict, loss, monitors, cmdviz_dict
        )

        if loss.requires_grad:
            self.trigger_event(
                "backward:before", self, feed_dict, loss, monitors, cmdviz_dict
            )
            # The design is useful for accumulated loss
            if self.loss_backward(loss):
                self.trigger_event(
                    "backward:after", self, feed_dict, loss, monitors, cmdviz_dict
                )
                self.optimizer_step()
        self.trigger_event("step:summary", self, loss, monitors, cmdviz_dict)

    def _eval_iter(self):
        self.eval()

    def _eval_epoch(self):
        self.eval()

    def optimizer_step(self):
        self.amp_scaler.step(self.optimizer)
        self.amp_scaler.update()
        self.optimizer.zero_grad()

    def loss_backward(self, loss):
        self.amp_scaler.scale(loss).backward()
        if self.iter_cnt % self.ratio_forback == 0:
            # TODO: unscales the gradient
            self.amp_scaler.unscale_(self.optimizer)
            return True
        return False

    def monitor_update(self):
        pass

    def forward(self, *args, **kwargs):
        if self.model is not None:
            return self.model(*args, **kwargs)
        return None

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)



import torch
from torch import nn
import jamtorch.utils.ts_ops as ts_ops

def is_positive_int(x):
    return isinstance(x, int) and x > 0

class NoMeanException(Exception):
    """Exception to be thrown when a mean function doesn't exist."""

    pass


class Distribution(nn.Module):
    """Base class for all distribution objects."""

    def forward(self, *args):
        raise RuntimeError("Forward method cannot be called for a Distribution object.")

    def log_prob(self, inputs, context=None):
        """Calculate log probability under the distribution.

        Args:
            inputs: Tensor, input variables.
            context: Tensor or None, conditioning variables. If a Tensor, it must have the same
                number or rows as the inputs. If None, the context is ignored.

        Returns:
            A Tensor of shape [input_size], the log probability of the inputs given the context.
        """
        inputs = torch.as_tensor(inputs)
        if context is not None:
            context = torch.as_tensor(context)
            if inputs.shape[0] != context.shape[0]:
                raise ValueError(
                    "Number of input items must be equal to number of context items."
                )
        return self._log_prob(inputs, context)

    def _log_prob(self, inputs, context):
        raise NotImplementedError()

    def sample(self, num_samples, context=None, batch_size=None):
        """Generates samples from the distribution. Samples can be generated in batches.

        Args:
            num_samples: int, number of samples to generate.
            context: Tensor or None, conditioning variables. If None, the context is ignored.
            batch_size: int or None, number of samples per batch. If None, all samples are generated
                in one batch.

        Returns:
            A Tensor containing the samples, with shape [num_samples, ...] if context is None, or
            [context_size, num_samples, ...] if context is given.
        """
        if not is_positive_int(num_samples):
            raise TypeError("Number of samples must be a positive integer.")

        if context is not None:
            context = torch.as_tensor(context)

        if batch_size is None:
            return self._sample(num_samples, context)

        else:
            if not is_positive_int(batch_size):
                raise TypeError("Batch size must be a positive integer.")

            num_batches = num_samples // batch_size
            num_leftover = num_samples % batch_size
            samples = [self._sample(batch_size, context) for _ in range(num_batches)]
            if num_leftover > 0:
                samples.append(self._sample(num_leftover, context))
            return torch.cat(samples, dim=0)

    def _sample(self, num_samples, context):
        raise NotImplementedError()

    def sample_and_log_prob(self, num_samples, context=None):
        """Generates samples from the distribution together with their log probability.

        Args:
            num_samples: int, number of samples to generate.
            context: Tensor or None, conditioning variables. If None, the context is ignored.

        Returns:
            A tuple of:
                * A Tensor containing the samples, with shape [num_samples, ...] if context is None,
                  or [context_size, num_samples, ...] if context is given.
                * A Tensor containing the log probabilities of the samples, with shape
                  [num_samples, ...] if context is None, or [context_size, num_samples, ...] if
                  context is given.
        """
        samples = self.sample(num_samples, context=context)

        if context is not None:
            # Merge the context dimension with sample dimension in order to call log_prob.
            samples = ts_ops.merge_leading_dims(samples, num_dims=2)
            context = ts_ops.repeat_rows(context, num_reps=num_samples)
            assert samples.shape[0] == context.shape[0]

        log_prob = self.log_prob(samples, context=context)

        if context is not None:
            # Split the context dimension from sample dimension.
            samples = ts_ops.split_leading_dim(samples, shape=[-1, num_samples])
            log_prob = ts_ops.split_leading_dim(log_prob, shape=[-1, num_samples])

        return samples, log_prob

    def mean(self, context=None):
        if context is not None:
            context = torch.as_tensor(context)
        return self._mean(context)

    def _mean(self, context):
        raise NoMeanException()



from typing import Any, Dict, List, Tuple, Union

import numpy as np
import torch as th

from jammy.utils.registry import CallbackRegistry

__all__ = ["merge", "merge_tree"]


def merge_dict(obj: List[Dict]):
    first_item = obj[0]
    rtn_dict = {}
    for key in first_item.keys():
        items = [cur_obj[key] for cur_obj in obj]
        rtn_dict[key] = merge(items)
    return rtn_dict


def merge_list(obj: List[List]):
    assert type(obj[0]) in [list]
    rtn = []
    for cur_list in obj:
        rtn.extend(cur_list)
    return rtn


def merge_basic(obj: List[Union[int, float, str, complex]]):
    assert type(obj[0]) in [int, float, str, complex]
    return obj


def merge_tuple(obj: List[Tuple]):
    length = len(obj[0])
    rtn = []
    for cur_idx in range(length):
        cur_items = [cur_obj[cur_idx] for cur_obj in obj]
        rtn.append(merge(cur_items))
    return tuple(rtn)


def merge_ndarray(obj: List[np.ndarray]):
    return np.concatenate(obj)


def merge_thtensor(obj: List[th.Tensor]):
    return th.cat(obj)


def _default_pytree_fallback(obj, *args, **kwargs):
    del args, kwargs
    if len(obj) == 0:
        return obj
    item_type = type(obj[0])
    raise ValueError('Unknown itme type: "{}".'.format(item_type))


class _PyTreeCallbackRegistry(CallbackRegistry):
    def dispatch(self, name, *args, **kwargs):
        return super().dispatch_direct(name, *args, **kwargs)


_pytree_registry = _PyTreeCallbackRegistry()
_pytree_registry.set_fallback_callback(_default_pytree_fallback)
_pytree_registry.register(int, merge_basic)
_pytree_registry.register(float, merge_basic)
_pytree_registry.register(complex, merge_basic)
_pytree_registry.register(np.ndarray, merge_ndarray)
_pytree_registry.register(th.Tensor, merge_thtensor)
_pytree_registry.register(tuple, merge_tuple)
_pytree_registry.register(dict, merge_dict)
_pytree_registry.register(list, merge_list)


def merge(obj: List[Any]):
    """
    List can only be in the leaf
    """
    if len(obj) == 0:
        return []
    item = obj[0]
    return _pytree_registry.dispatch(type(item), obj)


merge_tree = merge

if __name__ == "__main__":
    from jammy.utils.printing import stprint

    a = (1, np.arange(3), th.arange(4), {"a": 8})
    b = (2, np.arange(3) + 1, th.arange(4) + 5, {"a": 8})
    stprint(merge([a, b]))



import argparse
import os
import time
import uuid

import gpustat
import timeout_decorator

from jammy.comm import get_local_addr, is_port_used
from jammy.comm.cs import ClientPipe, ServerPipe
from jammy.logging import get_logger
from jammy.utils import gpu
from jammy.utils.env import jam_getenv
from jammy.utils.printing import kvformat

logger = get_logger()


def req_util(pipe, identifier, inp=None):
    num_gpus = inp.get("num_gpus", 1)
    sleep_sec = inp.get("sleep_sec", 3)
    mem_prior = inp.get("mem_prior", 0.5)
    idx = identifier.decode("ascii")
    logger.info(f"REQ from {idx}\n {kvformat(inp)}")
    gpu_list = gpu.gpu_by_weight(mem_prior)
    msg = None
    if not isinstance(gpu_list, list):
        gpu_list = [gpu_list]
    if len(gpu_list) < num_gpus:
        msg = gpu_list
    else:
        msg = gpu_list[:num_gpus]
    pipe.send(identifier, msg)
    logger.info(f"SEND: {idx} {msg}")
    time.sleep(sleep_sec)
    logger.info(f"CLOSE: {idx}")


def len_available_gpu(pipe, identifier, inp=None):
    del inp
    pipe.send(identifier, len(gpustat.new_query()))


def start_sever():
    parser = argparse.ArgumentParser(
        prog="jgpus: start a gpu resource allocation server"
    )
    parser.add_argument("-p", "--port", type=int, default=-1, help="set default port")
    args = parser.parse_args()

    server = ServerPipe("server", mode="tcp")
    server.dispatcher.register("req_util", req_util)
    server.dispatcher.register("total", len_available_gpu)
    if args.port < 0:
        p_router = jam_getenv("GPUPort", default=1080, type=int)
        logger.warning(f"Use default port {p_router}")
    else:
        p_router = args.port
        logger.warning(f"Use passed Port {p_router}")
    assert not is_port_used(p_router)
    assert not is_port_used(p_router + 1)
    with server.activate(tcp_port=[str(p_router), str(p_router + 1)]):
        logger.critical("GPU ROUTER and PULL IP:")
        logger.critical(server.conn_info)
        while True:
            time.sleep(1)


def instantiate_client(flag="_", port=-1):
    p_dealer = port
    if port < 0:
        p_dealer = jam_getenv("GPUPort", default=1080, type=int)
    tcp_dealer = f"tcp://{get_local_addr()}:{p_dealer}"
    tcp_push = f"tcp://{get_local_addr()}:{p_dealer+1}"
    client = ClientPipe(
        f"gpu_client{flag}" + uuid.uuid4().hex[:8], conn_info=[tcp_dealer, tcp_push]
    )
    return client


@timeout_decorator.timeout(
    5, exception_message="Make sure jgpus(gpu-server) has started"
)
def start_client():
    parser = argparse.ArgumentParser(
        prog="jgpuc: start a gpu resource allocation client"
    )
    parser.add_argument("-p", "--port", type=int, default=-1, help="set default port")
    args = parser.parse_args()

    client = instantiate_client(port=args.port)
    logger.info("Identity: {}.".format(client.identity))
    with client.activate():
        echo = client.query("total")
        logger.info(f"Contains {echo} gpus")
        query_gpu_ids = client.query("req_util", dict(num_gpus=1))
        logger.info(f"req get {query_gpu_ids}")


def get_gpu_by_utils(
    num_gpus: int = 1,
    sleep_sec: int = 3,
    timeout_sec: int = 7,
    mem_prior: float = 0.5,
    port: int = -1,
):
    @timeout_decorator.timeout(
        timeout_sec, exception_message="Make sure jgpu-server has started"
    )
    def work_fn():
        client = instantiate_client(os.getpid(), port=port)
        with client.activate():
            inp = {
                "num_gpus": num_gpus,
                "sleep_sec": sleep_sec,
                "mem_prior": mem_prior,
            }
            query_gpu_ids = client.query("req_util", inp)
        return query_gpu_ids

    return work_fn()
